"ooolib-python - Copyright (C) 2006-2007 Joseph Colton"

# ooolib-python - Python module for creating Open Document Format documents.
# Copyright (C) 2006-2007  Joseph Colton

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

# You can contact me by email at josephcolton@gmail.com

# Import Standard Modules
import zipfile
import time
import sys
import glob
import os
import re

def version():
    "Get the ooolib-python version"
    return "ooolib-python-0.0.11"

def clean_string(data):
    "Returns an XML friendly copy of the data string"
    data = str(data)
    data = data.replace('&', '&amp;')
    data = data.replace("'", '&apos;')
    data = data.replace('"', '&quot;')
    data = data.replace('<', '&lt;')
    data = data.replace('>', '&gt;')
    data = data.replace('\t', '<text:tab-stop/>')
    data = data.replace('\n', '<text:line-break/>')
    return data

class XML:
    "XML Class - Used to convert nested lists into XML"
    def __init__(self):
        "Initialize ooolib XML instance"
        pass

    def _xmldata(self, data):
        datatype = data.pop(0)
        datavalue = data.pop(0)
        outstring = '%s' % datavalue
        return outstring

    def _xmltag(self, data):
        outstring = ''
        # First two
        datatype = data.pop(0)
        dataname = data.pop(0)
        outstring = '<%s' % dataname
        # Element Section
        element = 1
        while(data):
            # elements
            newdata = data.pop(0)
            if (newdata[0] == 'element' and element):
                newstring = self._xmlelement(newdata)
                outstring = '%s %s' % (outstring, newstring)
                continue
            if (newdata[0] != 'element' and element):
                element = 0
                outstring = '%s>' % outstring
                if (newdata[0] == 'tag' or newdata[0] == 'tagline'):
                    outstring = '%s\n' % outstring
            if (newdata[0] == 'tag'):
                newstring = self._xmltag(newdata)
                outstring = '%s%s' % (outstring, newstring)
                continue
            if (newdata[0] == 'tagline'):
                newstring = self._xmltagline(newdata)
                outstring = '%s%s' % (outstring, newstring)
                continue
            if (newdata[0] == 'data'):
                newstring = self._xmldata(newdata)
                outstring = '%s%s' % (outstring, newstring)
                continue
        if (element):
            element = 0
            outstring = '%s>\n' % outstring
        outstring = '%s</%s>\n' % (outstring, dataname)
        return outstring

    def _xmltagline(self, data):
        outstring = ''
        # First two
        datatype = data.pop(0)
        dataname = data.pop(0)
        outstring = '<%s' % dataname
        # Element Section
        while(data):
            # elements
            newdata = data.pop(0)
            if (newdata[0] != 'element'): break
            newstring = self._xmlelement(newdata)
            outstring = '%s %s' % (outstring, newstring)
        outstring = '%s/>\n' % outstring
        # Non-Element Section should not exist
        return outstring

    def _xmlelement(self, data):
        datatype = data.pop(0)
        dataname = data.pop(0)
        datavalue = data.pop(0)
        outstring = '%s="%s"' % (dataname, datavalue)
        return outstring

    def convert(self, data):
        """Convert nested lists into XML

        The convert method takes a nested lists and converts them
        into XML to be used in Open Document Format documents.
        There are three types of lists that are recognized at this
        time.  They are as follows:

        'tag' - Tag opens a set of data that is eventually closed
        with a similar tag.
        List: ['tag', 'xml']
        XML: <xml></xml>

        'tagline' - Taglines are similar to tags, except they open
        and close themselves.
        List: ['tagline', 'xml']
        XML: <xml/>

        'element' - Elements are pieces of information stored in an
        opening tag or tagline.
        List: ['element', 'color', 'blue']
        XML: color="blue"

        'data' - Data is plain text directly inserted into the XML
        document.
        List: ['data', 'hello']
        XML: hello

        Bring them all together for something like this.

        Lists:
        ['tag', 'xml', ['element', 'a', 'b'], ['tagline', 'xml2'],
        ['data', 'asdf']]

        XML:
        <xml a="b"><xml2/>asdf</xml>
        """
        outlines = []
        outlines.append('<?xml version="1.0" encoding="UTF-8"?>')
        if (type(data) == type([]) and len(data) > 0):
            if data[0] == 'tag': outlines.append(self._xmltag(data))
        return outlines

class Meta:
    "Meta Data Class"

    def __init__(self, doctype):
        self.doctype = doctype
        self.meta_generator = version()       # The generator should always default to the version number
        self.meta_title = ''
        self.meta_subject = ''
        self.meta_description = ''
        self.meta_keywords = []
        self.meta_creator = 'ooolib-python'
        self.meta_editor = ''
        self.meta_user1_name = 'Info 1'
        self.meta_user2_name = 'Info 2'
        self.meta_user3_name = 'Info 3'
        self.meta_user4_name = 'Info 4'
        self.meta_user1_value = ''
        self.meta_user2_value = ''
        self.meta_user3_value = ''
        self.meta_user4_value = ''
        self.meta_creation_date = self.meta_time()

    def set_meta(self, metaname, value):
        """Set meta data in your document.

        Currently implemented metaname options are as follows:
        'creator' - The document author
        """
        if metaname == 'creator': self.meta_creator = value
        if metaname == 'editor': self.meta_editor = value
        if metaname == 'title': self.meta_title = value
        if metaname == 'subject': self.meta_subject = value
        if metaname == 'description': self.meta_description = value
        if metaname == 'user1name': self.meta_user1_name = value
        if metaname == 'user2name': self.meta_user2_name = value
        if metaname == 'user3name': self.meta_user3_name = value
        if metaname == 'user4name': self.meta_user4_name = value
        if metaname == 'user1value': self.meta_user1_value = value
        if metaname == 'user2value': self.meta_user2_value = value
        if metaname == 'user3value': self.meta_user3_value = value
        if metaname == 'user4value': self.meta_user4_value = value
        if metaname == 'keyword':
            if value not in self.meta_keywords:
                self.meta_keywords.append(value)

    def meta_time(self):
        "Return time string in meta data format"
        t = time.localtime()
        stamp = "%04d-%02d-%02dT%02d:%02d:%02d" % (t[0], t[1], t[2], t[3], t[4], t[5])
        return stamp

    def get_meta(self):
        "Generate meta.xml file data"
        self.meta_date = self.meta_time()
        self.data = ['tag', 'office:document-meta',
          ['element', 'xmlns:office', 'urn:oasis:names:tc:opendocument:xmlns:office:1.0'],
          ['element', 'xmlns:xlink', 'http://www.w3.org/1999/xlink'],
          ['element', 'xmlns:dc', 'http://purl.org/dc/elements/1.1/'],
          ['element', 'xmlns:meta', 'urn:oasis:names:tc:opendocument:xmlns:meta:1.0'],
          ['element', 'xmlns:ooo', 'http://openoffice.org/2004/office'],
          ['element', 'office:version', '1.0'],
          ['tag', 'office:meta',
            ['tag', 'meta:generator',             # Was: 'OpenOffice.org/2.0$Linux OpenOffice.org_project/680m5$Build-9011'
              ['data', self.meta_generator]],     # Generator is set the the ooolib-python version.
            ['tag', 'dc:title',
              ['data', self.meta_title]],         # This data is the document title
            ['tag', 'dc:description',
              ['data', self.meta_description]],   # This data is the document description
            ['tag', 'dc:subject',
              ['data', self.meta_subject]],       # This data is the document subject
            ['tag', 'meta:initial-creator',
              ['data', self.meta_creator]],       # This data is the document creator
            ['tag', 'meta:creation-date',
              ['data', self.meta_creation_date]], # This is the original creation date of the document
            ['tag', 'dc:creator',
              ['data', self.meta_editor]],        # This data is the document editor
            ['tag', 'dc:date',
              ['data', self.meta_date]],          # This is the last modified date of the document
            ['tag', 'dc:language',
              ['data', 'en-US']],                 # We will probably always use en-US for language
            ['tag', 'meta:editing-cycles',
              ['data', '1']],                     # Edit cycles will probably always be 1 for generated documents
            ['tag', 'meta:editing-duration',
              ['data', 'PT0S']],                  # Editing duration is modified - creation date
            ['tag', 'meta:user-defined',
              ['element', 'meta:name', self.meta_user1_name],
              ['data', self.meta_user1_value]],
            ['tag', 'meta:user-defined',
              ['element', 'meta:name', self.meta_user2_name],
              ['data', self.meta_user2_value]],
            ['tag', 'meta:user-defined',
              ['element', 'meta:name', self.meta_user3_name],
              ['data', self.meta_user3_value]],
            ['tag', 'meta:user-defined',
              ['element', 'meta:name', self.meta_user4_name],
              ['data', self.meta_user4_value]]]]
#                   ['tagline', 'meta:document-statistic',
#                     ['element', 'meta:table-count', len(self.sheets)], # len(self.sheets) ?
#                     ['element', 'meta:cell-count', '15']]]] # Not sure how to keep track

        # Generate content.xml XML data
        xml = XML()
        self.lines = xml.convert(self.data)
        self.filedata = '\n'.join(self.lines)
        # Return generated data
        return self.filedata



class CalcStyles:
    "Calc Style Management - Used to keep track of created styles."

    def __init__(self):
        self.style_config = {}
        # Style Counters
        self.style_table = 1
        self.style_column = 1
        self.style_row = 1
        self.style_cell = 1
        # Style Properties (Defaults) - To be used later
        self.property_column_width_default = '0.8925in' # Default Column Width
        self.property_row_height_default = '0.189in'    # Default Row Height
        # Set Defaults
        self.property_column_width = '0.8925in' # Default Column Width
        self.property_row_height = '0.189in'    # Default Row Height
        self.property_cell_bold = False         # Bold off be default
        self.property_cell_italic = False       # Italic off be default
        self.property_cell_underline = False    # Underline off be default
        self.property_cell_fg_color = 'default' # Text Color Default
        self.property_cell_bg_color = 'default' # Cell Background Default
        self.property_cell_fontsize = '10'      # Cell Font Size Default
        self.property_cell_valign = 'default'   # Vertial Alignment Default
        self.property_cell_halign = 'default'   # Horizantal Alignment Default

    def get_next_style(self, style):
        "Returns the next style code for the given style"
        style_code = ""
        if style == 'table':
            style_code = 'ta%d' % self.style_table
            self.style_table+=1
        if style == 'column':
            style_code = 'co%d' % self.style_column
            self.style_column+=1
        if style == 'row':
            style_code = 'ro%d' % self.style_row
            self.style_row+=1
        if style == 'cell':
            style_code = 'ce%d' % self.style_cell
            self.style_cell+=1
        return style_code

    def set_property(self, style, name, value):
        "Sets a property which will later be turned into a code"
        if style == 'table':
            pass
        if style == 'column':
            if name == 'style:column-width': self.property_column_width = value
        if style == 'row':
            if name == 'style:row-height': self.property_row_height = value
        if style == 'cell':
            if name == 'bold' and type(value) == type(True): self.property_cell_bold = value
            if name == 'italic' and type(value) == type(True): self.property_cell_italic = value
            if name == 'underline' and type(value) == type(True): self.property_cell_underline = value
            if name == 'fontsize': self.property_cell_fontsize = value
            if name == 'color':
                self.property_cell_fg_color = 'default'
                redata = re.search("^(#[\da-fA-F]{6})$", value)
                if redata: self.property_cell_fg_color = value.lower()
            if name == 'background':
                self.property_cell_bg_color = 'default'
                redata = re.search("^(#[\da-fA-F]{6})$", value)
                if redata: self.property_cell_bg_color = value.lower()
            if name == 'valign':
                self.property_cell_valign = value
            if name == 'halign':
                self.property_cell_halign = value

    def get_style_code(self, style):
        style_code = ""
        if style == 'table':
            style_code = "ta1"
        if style == 'column':
            style_data = tuple([style,
                    ('style:column-width', self.property_column_width)])
            if style_data in self.style_config:
                # Style Exists, return code
                style_code = self.style_config[style_data]
            else:
                # Style does not exist, create code and return it
                style_code = self.get_next_style(style)
                self.style_config[style_data] = style_code
        if style == 'row':
            style_data = tuple([style,
                    ('style:row-height', self.property_row_height)])
            if style_data in self.style_config:
                # Style Exists, return code
                style_code = self.style_config[style_data]
            else:
                # Style does not exist, create code and return it
                style_code = self.get_next_style(style)
                self.style_config[style_data] = style_code
        if style == 'cell':
            style_data = [style]
            # Add additional styles
            if self.property_cell_bold: style_data.append(('bold', True))
            if self.property_cell_italic: style_data.append(('italic', True))
            if self.property_cell_underline: style_data.append(('underline', True))
            if self.property_cell_fontsize != '10': style_data.append(('fontsize', self.property_cell_fontsize))
            if self.property_cell_fg_color != 'default': style_data.append(('color', self.property_cell_fg_color))
            if self.property_cell_bg_color != 'default': style_data.append(('background', self.property_cell_bg_color))
            if self.property_cell_valign != 'default': style_data.append(('valign', self.property_cell_valign))
            if self.property_cell_halign != 'default': style_data.append(('halign', self.property_cell_halign))

            style_data = tuple(style_data)
            if style_data in self.style_config:
                # Style Exists, return code
                style_code = self.style_config[style_data]
            else:
                # Style does not exist, create code and return it
                style_code = self.get_next_style(style)
                self.style_config[style_data] = style_code
        return style_code

    def get_automatic_styles(self):
        "Return 'office:automatic-styles' lists"
        automatic_styles = ['tag', 'office:automatic-styles']

        for style_data in self.style_config:
            style_code = self.style_config[style_data]
            style_data = list(style_data)
            style = style_data.pop(0)

            if style == 'column':
                style_list = ['tag', 'style:style',
                  ['element', 'style:name', style_code],            # Column 'co1' properties
                  ['element', 'style:family', 'table-column']]
                tagline = ['tagline', 'style:table-column-properties',
                  ['element', 'fo:break-before', 'auto']]      # unsure what break before means

                for set in style_data:
                    name, value = set
                    if name == 'style:column-width':
                        tagline.append(['element', 'style:column-width', value])
                style_list.append(tagline)
                automatic_styles.append(style_list)

            if style == 'row':
                style_list = ['tag', 'style:style',
                  ['element', 'style:name', style_code],            # Column 'ro1' properties
                  ['element', 'style:family', 'table-row']]
                tagline = ['tagline', 'style:table-row-properties']

                for set in style_data:
                    name, value = set
                    if name == 'style:row-height':
                        tagline.append(['element', 'style:row-height', value])
                tagline.append(['element', 'fo:break-before', 'auto'])
#                               tagline.append(['element', 'style:use-optimal-row-height', 'true']) # Overrides settings
                style_list.append(tagline)
                automatic_styles.append(style_list)

            if style == 'cell':
                style_list = ['tag', 'style:style',
                  ['element', 'style:name', style_code],             # ce1 style
                  ['element', 'style:family', 'table-cell'],         # cell
                  ['element', 'style:parent-style-name', 'Default']] # parent is Default

                # Cell Properties
                tagline = ['tagline', 'style:table-cell-properties']
                for set in style_data:
                    name, value = set
                    if name == 'background':
                        tagline.append(['element', 'fo:background-color', value])
                    if name == 'valign':
                        if value in ['top', 'bottom', 'middle']:
                            tagline.append(['element', 'style:vertical-align', value])
                    if name == 'halign':
                        tagline.append(['element', 'style:text-align-source', 'fix'])
                        if value in ['filled']:
                            tagline.append(['element', 'style:repeat-content', 'true'])
                        else:
                            tagline.append(['element', 'style:repeat-content', 'false'])

                style_list.append(tagline)

                # Paragraph Properties
                tagline = ['tagline', 'style:paragraph-properties']
                tagline_valid = False
                for set in style_data:
                    name, value = set
                    if name == 'halign':
                        tagline_valid = True
                        if value in ['center']:
                            tagline.append(['element', 'fo:text-align', 'center'])
                        if value in ['end', 'right']:
                            tagline.append(['element', 'fo:text-align', 'end'])
                        if value in ['start', 'filled', 'left']:
                            tagline.append(['element', 'fo:text-align', 'start'])
                        if value in ['justify']:
                            tagline.append(['element', 'fo:text-align', 'justify'])
                # Conditionally add the tagline
                if tagline_valid: style_list.append(tagline)


                # Text Properties
                tagline = ['tagline', 'style:text-properties']
                for set in style_data:
                    name, value = set
                    if name == 'bold':
                        tagline.append(['element', 'fo:font-weight', 'bold'])
                    if name == 'italic':
                        tagline.append(['element', 'fo:font-style', 'italic'])
                    if name == 'underline':
                        tagline.append(['element', 'style:text-underline-style', 'solid'])
                        tagline.append(['element', 'style:text-underline-width', 'auto'])
                        tagline.append(['element', 'style:text-underline-color', 'font-color'])
                    if name == 'color':
                        tagline.append(['element', 'fo:color', value])
                    if name == 'fontsize':
                        tagline.append(['element', 'fo:font-size', '%spt' % value])
                style_list.append(tagline)

                automatic_styles.append(style_list)


        # Attach ta1 style
        automatic_styles.append(['tag', 'style:style',
          ['element', 'style:name', 'ta1'],
          ['element', 'style:family', 'table'],
          ['element', 'style:master-page-name', 'Default'],
          ['tagline', 'style:table-properties',
            ['element', 'table:display', 'true'],
            ['element', 'style:writing-mode', 'lr-tb']]])


        return automatic_styles



class CalcSheet:
    "Calc Sheet Class - Used to keep track of the data for an individual sheet."

    def __init__(self, sheetname):
        "Initialize a sheet"
        self.sheet_name = sheetname
        self.sheet_values = {}
        self.sheet_config = {}
        self.max_col = 0
        self.max_row = 0

    def clean_formula(self, data):
        "Returns a formula for use in ODF"
        # Example Translations
        # '=SUM(A1:A2)'
        # datavalue = 'oooc:=SUM([.A1:.A2])'
        data = str(data)
        redata = re.search('^=([A-Z]+)\(([A-Z0-9:]+)\)$', data)
        if redata:
            funct = redata.group(1)
            parts = redata.group(2).split(':')
            nparts = []
            while parts:
                part = parts.pop(0)
                part = '.%s' % part
                nparts.append(part)
            parts = ':'.join(nparts)
            data = 'oooc:=%s([%s])' % (funct, parts)
        return data

    def get_name(self):
        "Returns the sheet name"
        return self.sheet_name

    def get_sheet_values(self):
        "Returns the sheet cell values"
        return self.sheet_values

    def get_sheet_config(self):
        "Returns the sheet cell properties"
        return self.sheet_config

    def set_sheet_config(self, location, style_code):
        "Sets Style Code for a given location"
        self.sheet_config[location] = style_code

    def set_sheet_value(self, cell, datatype, datavalue):
        """Sets the value for a specific cell

        cell must be in the format (col, row) where row and col are int.
        Example: B5 would be written as (2, 5)
        datatype must be one of 'string', 'float', 'formula'
        datavalue should be a string
        """
        # Catch invalid data
        if type(cell) != type(()) or len(cell) != 2:
            print "Invalid Cell"
            return
        (col, row) = cell
        if type(col) != type(1):
            print "Invalid Cell"
            return
        if type(row) != type(1):
            print "Invalid Cell"
            return
        # Fix String Data
        if datatype in ['string', 'annotation']:
            datavalue = clean_string(datavalue)
        # Fix Link Data. Link's value is a tuple containing (url, description)
        if (datatype == 'link'):
            url = clean_string(datavalue[0])
            desc = clean_string(datavalue[1])
            datavalue = (url, desc)
        # Fix Formula Data
        if datatype == 'formula':
            datavalue = self.clean_formula(datavalue)
        # Adjust maximum sizes
        if col > self.max_col: self.max_col = col
        if row > self.max_row: self.max_row = row
        datatype = str(datatype)
        if (datatype not in ['string', 'float', 'formula', 'annotation', 'link']):
            # Set all unknown cell types to string
            datatype = 'string'
            datavalue = str(datavalue)

        # The following lines are taken directly from HPS
        # self.sheet_values[cell] = (datatype, datavalue)
        # HPS: Cell content is now a list of tuples instead of a tuple
        # While storing here, store the cell contents first and the annotation next. While generating the XML reverse this
        contents = self.sheet_values.get(cell, {'annotation':None,'link':None, 'value':None})
        if datatype == 'annotation':
            contents['annotation'] = (datatype, datavalue)
        elif datatype == 'link':
            contents['link'] = (datatype, datavalue)
        else:
            contents['value'] = (datatype, datavalue)

        self.sheet_values[cell] = contents


    def get_lists(self):
        "Returns nested lists for XML processing"
        if (self.max_col == 0 and self.max_row == 0):
            sheet_lists = ['tag', 'table:table',
              ['element', 'table:name', self.sheet_name], # Set the Sheet Name
              ['element', 'table:style-name', 'ta1'],
              ['element', 'table:print', 'false'],
              ['tagline', 'table:table-column',
                ['element', 'table:style-name', 'co1'],
                ['element', 'table:default-cell-style-name', 'Default']],
                ['tag', 'table:table-row',
                  ['element', 'table:style-name', 'ro1'],
                  ['tagline', 'table:table-cell']]]
        else:
            # Base Information
            sheet_lists = ['tag', 'table:table',
              ['element', 'table:name', self.sheet_name], # Set the sheet name
              ['element', 'table:style-name', 'ta1'],
              ['element', 'table:print', 'false']]

#                         ['tagline', 'table:table-column',
#                           ['element', 'table:style-name', 'co1'],
#                           ['element', 'table:number-columns-repeated', self.max_col], # max_col? '2'
#                           ['element', 'table:default-cell-style-name', 'Default']],

            # Need to add column information
            for col in range(1, self.max_col+1):
                location = ('col', col)
                style_code = 'co1'
                if location in self.sheet_config:
                    style_code = self.sheet_config[location]
                sheet_lists.append(['tagline', 'table:table-column',
                  ['element', 'table:style-name', style_code],
                  ['element', 'table:default-cell-style-name', 'Default']])


            # Need to create each row
            for row in range(1, self.max_row + 1):
                location = ('row', row)
                style_code = 'ro1'
                if location in self.sheet_config:
                    style_code = self.sheet_config[location]
                rowlist = ['tag', 'table:table-row',
                  ['element', 'table:style-name', style_code]]
                for col in range(1, self.max_col + 1):
                    cell = (col, row)
                    style_code = 'ce1'                           # Default all cells to ce1
                    if cell in self.sheet_config:
                        style_code = self.sheet_config[cell] # Lookup cell if available
                    if cell in self.sheet_values:
                        # (datatype, datavalue) = self.sheet_values[cell] # Marked for removal
                        collist = ['tag', 'table:table-cell']
                        if style_code != 'ce1':
                            collist.append(['element', 'table:style-name', style_code])

                        # Contents, annotations, and links added by HPS
                        contents = self.sheet_values[cell] # cell contents is a dictionary
                        if contents['value']:
                            (datatype, datavalue) = contents['value']
                            if datatype == 'float':
                                collist.append(['element', 'office:value-type', datatype])
                                collist.append(['element', 'office:value', datavalue])

                            if datatype == 'string':
                                collist.append(['element', 'office:value-type', datatype])
                            if datatype == 'formula':
                                collist.append(['element', 'table:formula', datavalue])
                                collist.append(['element', 'office:value-type', 'float'])
                                collist.append(['element', 'office:value', '0'])
                                datavalue = '0'
                        else:
                            datavalue = None

                        if contents['annotation']:
                            (annotype, annoval) = contents['annotation']
                            collist.append(['tag', 'office:annotation',
                              ['tag', 'text:p', ['data', annoval]]])

                        if contents['link']:
                            (linktype, linkval) = contents['link']
                            if datavalue:
                                collist.append(['tag', 'text:p', ['data', datavalue],
                                  ['tag', 'text:a', ['element', 'xlink:href', linkval[0]],
                                  ['data', linkval[1]]]])
                            else: # no value; just fill the link
                                collist.append(['tag', 'text:p',
                                  ['tag', 'text:a', ['element', 'xlink:href', linkval[0]],
                                  ['data', linkval[1]]]])
                        else:
                            if datavalue:
                                collist.append(['tag', 'text:p', ['data', datavalue]])



                    else:
                        collist = ['tagline', 'table:table-cell']
                    rowlist.append(collist)
                sheet_lists.append(rowlist)
        return sheet_lists

class Calc:
    "Calc Class - Used to create OpenDocument Format Calc Spreadsheets."
    def __init__(self, sheetname=None):
        "Initialize ooolib Calc instance"
        # Default to no debugging
        self.debug = False
        if not sheetname: sheetname = "Sheet1"
        self.sheets = [CalcSheet(sheetname)]  # The main sheet will be initially called 'Sheet1'
        self.sheet_index = 0                  # We initially start on the first sheet
        self.styles = CalcStyles()
        self.meta = Meta('ods')
        self.styles.get_style_code('column')  # Force generation of default column
        self.styles.get_style_code('row')     # Force generation of default row
        self.styles.get_style_code('table')   # Force generation of default table
        self.styles.get_style_code('cell')    # Force generation of default cell

    def debug_level(self, level):
        """Set debug level:
        True if you want debugging messages
        False if you do not.
        """
        self.debug = level

    def set_meta(self, metaname, value):
        "Set meta data in your document."
        self.meta.set_meta(metaname, value)

    def set_column_property(self, column, name, value):
        "Set Column Properties"
        if name == 'width':
            # column number column needs column-width set to value
            self.styles.set_property('column', 'style:column-width', value)
            style_code = self.styles.get_style_code('column')
            self.sheets[self.sheet_index].set_sheet_config(('col', column), style_code)

    def set_row_property(self, row, name, value):
        "Set row Properties"
        if name == 'height':
            # row number row needs row-height set to value
            self.styles.set_property('row', 'style:row-height', value)
            style_code = self.styles.get_style_code('row')
            self.sheets[self.sheet_index].set_sheet_config(('row', row), style_code)

    def set_cell_property(self, name, value):
        """Turn and off cell properties

        Actual application of properties is handled by setting a value."""
        self.styles.set_property('cell', name, value)

    def get_sheet_index(self):
        "Return the current sheet index number"
        return self.sheet_index

    def set_sheet_index(self, index):
        "Set the sheet index"
        if type(index) == type(1):
            if index >= 0 and index < len(self.sheets):
                self.sheet_index = index
        return self.sheet_index

    def new_sheet(self, sheetname):
        "Create a new sheet"
        self.sheet_index = len(self.sheets)
        self.sheets.append(CalcSheet(sheetname))
        return self.sheet_index

    def set_cell_value(self, col, row, datatype, value):
        "Set the value for a given cell"
        self.sheets[self.sheet_index].set_sheet_value((col, row), datatype, value)
        style_code = self.styles.get_style_code('cell')
        self.sheets[self.sheet_index].set_sheet_config((col, row), style_code)

    def save(self, filename):
        """Save .ods spreadsheet.

        The save function saves the current cells and settings into a document.
        """
        if self.debug: print "Writing %s" % filename
        self.savefile = zipfile.ZipFile(filename, "w")
        if self.debug: print "  meta.xml"
        self._zip_insert(self.savefile, "meta.xml", self.meta.get_meta())
        if self.debug: print "  mimetype"
        self._zip_insert(self.savefile, "mimetype", "application/vnd.oasis.opendocument.spreadsheet")
        if self.debug: print "  Configurations2/accelerator/current.xml"
        self._zip_insert(self.savefile, "Configurations2/accelerator/current.xml", "")
        if self.debug: print "  META-INF/manifest.xml"
        self._zip_insert(self.savefile, "META-INF/manifest.xml", self._ods_manifest())
        if self.debug: print "  content.xml"
        self._zip_insert(self.savefile, "content.xml", self._ods_content())
        if self.debug: print "  settings.xml"
        self._zip_insert(self.savefile, "settings.xml", self._ods_settings())
        if self.debug: print "  styles.xml"
        self._zip_insert(self.savefile, "styles.xml", self._ods_styles())

    def _zip_insert(self, file, filename, data):
        now = time.localtime(time.time())[:6]
        info = zipfile.ZipInfo(filename)
        info.date_time = now
        info.compress_type = zipfile.ZIP_DEFLATED
        file.writestr(info, data)

    def _ods_content(self):
        "Generate ods content.xml data"

        # This will list all of the sheets in the document
        self.sheetdata = ['tag', 'office:spreadsheet']
        for sheet in self.sheets:
            if self.debug:
                sheet_name = sheet.get_name()
                print "    Creating Sheet '%s'" % sheet_name
            sheet_list = sheet.get_lists()
            self.sheetdata.append(sheet_list)
        # Automatic Styles
        self.automatic_styles = self.styles.get_automatic_styles()

        self.data = ['tag', 'office:document-content',
          ['element', 'xmlns:office', 'urn:oasis:names:tc:opendocument:xmlns:office:1.0'],
          ['element', 'xmlns:style', 'urn:oasis:names:tc:opendocument:xmlns:style:1.0'],
          ['element', 'xmlns:text', 'urn:oasis:names:tc:opendocument:xmlns:text:1.0'],
          ['element', 'xmlns:table', 'urn:oasis:names:tc:opendocument:xmlns:table:1.0'],
          ['element', 'xmlns:draw', 'urn:oasis:names:tc:opendocument:xmlns:drawing:1.0'],
          ['element', 'xmlns:fo', 'urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0'],
          ['element', 'xmlns:xlink', 'http://www.w3.org/1999/xlink'],
          ['element', 'xmlns:dc', 'http://purl.org/dc/elements/1.1/'],
          ['element', 'xmlns:meta', 'urn:oasis:names:tc:opendocument:xmlns:meta:1.0'],
          ['element', 'xmlns:number', 'urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0'],
          ['element', 'xmlns:svg', 'urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0'],
          ['element', 'xmlns:chart', 'urn:oasis:names:tc:opendocument:xmlns:chart:1.0'],
          ['element', 'xmlns:dr3d', 'urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0'],
          ['element', 'xmlns:math', 'http://www.w3.org/1998/Math/MathML'],
          ['element', 'xmlns:form', 'urn:oasis:names:tc:opendocument:xmlns:form:1.0'],
          ['element', 'xmlns:script', 'urn:oasis:names:tc:opendocument:xmlns:script:1.0'],
          ['element', 'xmlns:ooo', 'http://openoffice.org/2004/office'],
          ['element', 'xmlns:ooow', 'http://openoffice.org/2004/writer'],
          ['element', 'xmlns:oooc', 'http://openoffice.org/2004/calc'],
          ['element', 'xmlns:dom', 'http://www.w3.org/2001/xml-events'],
          ['element', 'xmlns:xforms', 'http://www.w3.org/2002/xforms'],
          ['element', 'xmlns:xsd', 'http://www.w3.org/2001/XMLSchema'],
          ['element', 'xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance'],
          ['element', 'office:version', '1.0'],
          ['tagline', 'office:scripts'],
          ['tag', 'office:font-face-decls',
            ['tagline', 'style:font-face',
              ['element', 'style:name', 'DejaVu Sans'],
              ['element', 'svg:font-family', '&apos;DejaVu Sans&apos;'],
              ['element', 'style:font-pitch', 'variable']],
            ['tagline', 'style:font-face',
              ['element', 'style:name', 'Nimbus Sans L'],
              ['element', 'svg:font-family', '&apos;Nimbus Sans L&apos;'],
              ['element', 'style:font-family-generic', 'swiss'],
              ['element', 'style:font-pitch', 'variable']]],

        # Automatic Styles
        self.automatic_styles,

          ['tag', 'office:body',
            self.sheetdata]]                                      # Sheets are generated from the CalcSheet class

        # Generate content.xml XML data
        xml = XML()
        self.lines = xml.convert(self.data)
        self.filedata = '\n'.join(self.lines)
        # Return generated data
        return self.filedata

    def _ods_manifest(self):
        "Generate ods manifest.xml data"
        self.data = ['tag', 'manifest:manifest',
          ['element', 'xmlns:manifest', 'urn:oasis:names:tc:opendocument:xmlns:manifest:1.0'],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', 'application/vnd.oasis.opendocument.spreadsheet'],
            ['element', 'manifest:full-path', '/']],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', 'application/vnd.sun.xml.ui.configuration'],
            ['element', 'manifest:full-path', 'Configurations2/']],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', ''],
            ['element', 'manifest:full-path', 'Configurations2/accelerator/']],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', ''],
            ['element', 'manifest:full-path', 'Configurations2/accelerator/current.xml']],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', 'text/xml'],
            ['element', 'manifest:full-path', 'content.xml']],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', 'text/xml'],
            ['element', 'manifest:full-path', 'styles.xml']],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', 'text/xml'],
            ['element', 'manifest:full-path', 'meta.xml']],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', 'text/xml'],
            ['element', 'manifest:full-path', 'settings.xml']]]

        # Generate content.xml XML data
        xml = XML()
        self.lines = xml.convert(self.data)
        self.filedata = '\n'.join(self.lines)
        # Return generated data
        return self.filedata


    def _ods_settings(self):
        "Generate ods settings.xml data"
        self.data = ['tag', 'office:document-settings',
          ['element', 'xmlns:office', 'urn:oasis:names:tc:opendocument:xmlns:office:1.0'],
          ['element', 'xmlns:xlink', 'http://www.w3.org/1999/xlink'],
          ['element', 'xmlns:config', 'urn:oasis:names:tc:opendocument:xmlns:config:1.0'],
          ['element', 'xmlns:ooo', 'http://openoffice.org/2004/office'],
          ['element', 'office:version', '1.0'],
          ['tag', 'office:settings',
            ['tag', 'config:config-item-set',
              ['element', 'config:name', 'ooo:view-settings'],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'VisibleAreaTop'],
                ['element', 'config:type', 'int'],
                ['data', '0']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'VisibleAreaLeft'],
                ['element', 'config:type', 'int'],
                ['data', '0']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'VisibleAreaWidth'],
                ['element', 'config:type', 'int'],
                ['data', '6774']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'VisibleAreaHeight'],
                ['element', 'config:type', 'int'],
                ['data', '2389']],
              ['tag', 'config:config-item-map-indexed',
                ['element', 'config:name', 'Views'],
                ['tag', 'config:config-item-map-entry',
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'ViewId'],
                    ['element', 'config:type', 'string'],
                    ['data', 'View1']],
                  ['tag', 'config:config-item-map-named',
                    ['element', 'config:name', 'Tables'],
                    ['tag', 'config:config-item-map-entry',
                      ['element', 'config:name', 'Sheet1'],
                      ['tag', 'config:config-item',
                        ['element', 'config:name', 'CursorPositionX'], # Cursor Position A
                        ['element', 'config:type', 'int'],
                        ['data', '0']],
                      ['tag', 'config:config-item',
                        ['element', 'config:name', 'CursorPositionY'], # Cursor Position 1
                        ['element', 'config:type', 'int'],
                        ['data', '0']],
                      ['tag', 'config:config-item',
                        ['element', 'config:name', 'HorizontalSplitMode'],
                        ['element', 'config:type', 'short'],
                        ['data', '0']],
                      ['tag', 'config:config-item',
                        ['element', 'config:name', 'VerticalSplitMode'],
                        ['element', 'config:type', 'short'],
                        ['data', '0']],
                      ['tag', 'config:config-item',
                        ['element', 'config:name', 'HorizontalSplitPosition'],
                        ['element', 'config:type', 'int'],
                        ['data', '0']],
                      ['tag', 'config:config-item',
                        ['element', 'config:name', 'VerticalSplitPosition'],
                        ['element', 'config:type', 'int'],
                        ['data', '0']],
                      ['tag', 'config:config-item',
                        ['element', 'config:name', 'ActiveSplitRange'],
                        ['element', 'config:type', 'short'],
                        ['data', '2']],
                      ['tag', 'config:config-item',
                        ['element', 'config:name', 'PositionLeft'],
                        ['element', 'config:type', 'int'],
                        ['data', '0']],
                      ['tag', 'config:config-item',
                        ['element', 'config:name', 'PositionRight'],
                        ['element', 'config:type', 'int'],
                        ['data', '0']],
                      ['tag', 'config:config-item',
                        ['element', 'config:name', 'PositionTop'],
                        ['element', 'config:type', 'int'],
                        ['data', '0']],
                      ['tag', 'config:config-item',
                        ['element', 'config:name', 'PositionBottom'],
                        ['element', 'config:type', 'int'],
                        ['data', '0']]]],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'ActiveTable'],
                    ['element', 'config:type', 'string'],
                    ['data', 'Sheet1']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'HorizontalScrollbarWidth'],
                    ['element', 'config:type', 'int'],
                    ['data', '270']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'ZoomType'],
                    ['element', 'config:type', 'short'],
                    ['data', '0']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'ZoomValue'],
                    ['element', 'config:type', 'int'],
                    ['data', '100']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'PageViewZoomValue'],
                    ['element', 'config:type', 'int'],
                    ['data', '60']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'ShowPageBreakPreview'],
                    ['element', 'config:type', 'boolean'],
                    ['data', 'false']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'ShowZeroValues'],
                    ['element', 'config:type', 'boolean'],
                    ['data', 'true']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'ShowNotes'],
                    ['element', 'config:type', 'boolean'],
                    ['data', 'true']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'ShowGrid'],
                    ['element', 'config:type', 'boolean'],
                    ['data', 'true']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'GridColor'],
                    ['element', 'config:type', 'long'],
                    ['data', '12632256']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'ShowPageBreaks'],
                    ['element', 'config:type', 'boolean'],
                    ['data', 'true']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'HasColumnRowHeaders'],
                    ['element', 'config:type', 'boolean'],
                    ['data', 'true']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'HasSheetTabs'],
                    ['element', 'config:type', 'boolean'],
                    ['data', 'true']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'IsOutlineSymbolsSet'],
                    ['element', 'config:type', 'boolean'],
                    ['data', 'true']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'IsSnapToRaster'],
                    ['element', 'config:type', 'boolean'],
                    ['data', 'false']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'RasterIsVisible'],
                    ['element', 'config:type', 'boolean'],
                    ['data', 'false']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'RasterResolutionX'],
                    ['element', 'config:type', 'int'],
                    ['data', '1270']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'RasterResolutionY'],
                    ['element', 'config:type', 'int'],
                    ['data', '1270']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'RasterSubdivisionX'],
                    ['element', 'config:type', 'int'],
                    ['data', '1']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'RasterSubdivisionY'],
                    ['element', 'config:type', 'int'],
                    ['data', '1']],
                  ['tag', 'config:config-item',
                    ['element', 'config:name', 'IsRasterAxisSynchronized'],
                    ['element', 'config:type', 'boolean'],
                    ['data', 'true']]]]],
            ['tag', 'config:config-item-set',
              ['element', 'config:name', 'ooo:configuration-settings'],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'ShowZeroValues'],
                ['element', 'config:type', 'boolean'],
                ['data', 'true']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'ShowNotes'],
                ['element', 'config:type', 'boolean'],
                ['data', 'true']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'ShowGrid'],
                ['element', 'config:type', 'boolean'],
                ['data', 'true']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'GridColor'],
                ['element', 'config:type', 'long'],
                ['data', '12632256']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'ShowPageBreaks'],
                ['element', 'config:type', 'boolean'],
                ['data', 'true']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'LinkUpdateMode'],
                ['element', 'config:type', 'short'],
                ['data', '3']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'HasColumnRowHeaders'],
                ['element', 'config:type', 'boolean'],
                ['data', 'true']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'HasSheetTabs'],
                ['element', 'config:type', 'boolean'],
                ['data', 'true']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'IsOutlineSymbolsSet'],
                ['element', 'config:type', 'boolean'],
                ['data', 'true']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'IsSnapToRaster'],
                ['element', 'config:type', 'boolean'],
                ['data', 'false']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'RasterIsVisible'],
                ['element', 'config:type', 'boolean'],
                ['data', 'false']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'RasterResolutionX'],
                ['element', 'config:type', 'int'],
                ['data', '1270']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'RasterResolutionY'],
                ['element', 'config:type', 'int'],
                ['data', '1270']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'RasterSubdivisionX'],
                ['element', 'config:type', 'int'],
                ['data', '1']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'RasterSubdivisionY'],
                ['element', 'config:type', 'int'],
                ['data', '1']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'IsRasterAxisSynchronized'],
                ['element', 'config:type', 'boolean'],
                ['data', 'true']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'AutoCalculate'],
                ['element', 'config:type', 'boolean'],
                ['data', 'true']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'PrinterName'],
                ['element', 'config:type', 'string'],
                ['data', 'Generic Printer']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'PrinterSetup'],
                ['element', 'config:type', 'base64Binary'],
                ['data', 'YgH+/0dlbmVyaWMgUHJpbnRlcgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAU0dFTlBSVAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAWAAMAqAAAAAAA//8FAFZUAAAkbQAASm9iRGF0YSAxCnByaW50ZXI9R2VuZXJpYyBQcmludGVyCm9yaWVudGF0aW9uPVBvcnRyYWl0CmNvcGllcz0xCnNjYWxlPTEwMAptYXJnaW5kYWp1c3RtZW50PTAsMCwwLDAKY29sb3JkZXB0aD0yNApwc2xldmVsPTAKY29sb3JkZXZpY2U9MApQUERDb250ZXhEYXRhClBhZ2VTaXplOkxldHRlcgAA']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'ApplyUserData'],
                ['element', 'config:type', 'boolean'],
                ['data', 'true']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'CharacterCompressionType'],
                ['element', 'config:type', 'short'],
                ['data', '0']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'IsKernAsianPunctuation'],
                ['element', 'config:type', 'boolean'],
                ['data', 'false']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'SaveVersionOnClose'],
                ['element', 'config:type', 'boolean'],
                ['data', 'false']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'UpdateFromTemplate'],
                ['element', 'config:type', 'boolean'],
                ['data', 'false']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'AllowPrintJobCancel'],
                ['element', 'config:type', 'boolean'],
                ['data', 'true']],
              ['tag', 'config:config-item',
                ['element', 'config:name', 'LoadReadonly'],
                ['element', 'config:type', 'boolean'],
                ['data', 'false']]]]]

        # Generate content.xml XML data
        xml = XML()
        self.lines = xml.convert(self.data)
        self.filedata = '\n'.join(self.lines)
        # Return generated data
        return self.filedata


    def _ods_styles(self):
        "Generate ods styles.xml data"
        self.data = ['tag', 'office:document-styles',
          ['element', 'xmlns:office', 'urn:oasis:names:tc:opendocument:xmlns:office:1.0'],
          ['element', 'xmlns:style', 'urn:oasis:names:tc:opendocument:xmlns:style:1.0'],
          ['element', 'xmlns:text', 'urn:oasis:names:tc:opendocument:xmlns:text:1.0'],
          ['element', 'xmlns:table', 'urn:oasis:names:tc:opendocument:xmlns:table:1.0'],
          ['element', 'xmlns:draw', 'urn:oasis:names:tc:opendocument:xmlns:drawing:1.0'],
          ['element', 'xmlns:fo', 'urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0'],
          ['element', 'xmlns:xlink', 'http://www.w3.org/1999/xlink'],
          ['element', 'xmlns:dc', 'http://purl.org/dc/elements/1.1/'],
          ['element', 'xmlns:meta', 'urn:oasis:names:tc:opendocument:xmlns:meta:1.0'],
          ['element', 'xmlns:number', 'urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0'],
          ['element', 'xmlns:svg', 'urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0'],
          ['element', 'xmlns:chart', 'urn:oasis:names:tc:opendocument:xmlns:chart:1.0'],
          ['element', 'xmlns:dr3d', 'urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0'],
          ['element', 'xmlns:math', 'http://www.w3.org/1998/Math/MathML'],
          ['element', 'xmlns:form', 'urn:oasis:names:tc:opendocument:xmlns:form:1.0'],
          ['element', 'xmlns:script', 'urn:oasis:names:tc:opendocument:xmlns:script:1.0'],
          ['element', 'xmlns:ooo', 'http://openoffice.org/2004/office'],
          ['element', 'xmlns:ooow', 'http://openoffice.org/2004/writer'],
          ['element', 'xmlns:oooc', 'http://openoffice.org/2004/calc'],
          ['element', 'xmlns:dom', 'http://www.w3.org/2001/xml-events'],
          ['element', 'office:version', '1.0'],
          ['tag', 'office:font-face-decls',
            ['tagline', 'style:font-face',
              ['element', 'style:name', 'DejaVu Sans'],
              ['element', 'svg:font-family', '&apos;DejaVu Sans&apos;'],
              ['element', 'style:font-pitch', 'variable']],
            ['tagline', 'style:font-face',
              ['element', 'style:name', 'Nimbus Sans L'],
              ['element', 'svg:font-family', '&apos;Nimbus Sans L&apos;'],
              ['element', 'style:font-family-generic', 'swiss'],
              ['element', 'style:font-pitch', 'variable']]],
          ['tag', 'office:styles',
            ['tag', 'style:default-style',
              ['element', 'style:family', 'table-cell'],
              ['tagline', 'style:table-cell-properties',
                ['element', 'style:decimal-places', '2']],
              ['tagline', 'style:paragraph-properties',
                ['element', 'style:tab-stop-distance', '0.5in']],
              ['tagline', 'style:text-properties',
                ['element', 'style:font-name', 'Nimbus Sans L'],
                ['element', 'fo:language', 'en'],
                ['element', 'fo:country', 'US'],
                ['element', 'style:font-name-asian', 'DejaVu Sans'],
                ['element', 'style:language-asian', 'none'],
                ['element', 'style:country-asian', 'none'],
                ['element', 'style:font-name-complex', 'DejaVu Sans'],
                ['element', 'style:language-complex', 'none'],
                ['element', 'style:country-complex', 'none']]],
            ['tag', 'number:number-style',
              ['element', 'style:name', 'N0'],
              ['tagline', 'number:number',
                ['element', 'number:min-integer-digits', '1']]],
            ['tag', 'number:currency-style',
              ['element', 'style:name', 'N104P0'],
              ['element', 'style:volatile', 'true'],
              ['tag', 'number:currency-symbol',
                ['element', 'number:language', 'en'],
                ['element', 'number:country', 'US'],
                ['data', '$']],
              ['tagline', 'number:number',
                ['element', 'number:decimal-places', '2'],
                ['element', 'number:min-integer-digits', '1'],
                ['element', 'number:grouping', 'true']]],
            ['tag', 'number:currency-style',
              ['element', 'style:name', 'N104'],
              ['tagline', 'style:text-properties',
                ['element', 'fo:color', '#ff0000']],
              ['tag', 'number:text',
                ['data', '-']],
              ['tag', 'number:currency-symbol',
                ['element', 'number:language', 'en'],
                ['element', 'number:country', 'US'],
                ['data', '$']],
              ['tagline', 'number:number',
                ['element', 'number:decimal-places', '2'],
                ['element', 'number:min-integer-digits', '1'],
                ['element', 'number:grouping', 'true']],
              ['tagline', 'style:map',
                ['element', 'style:condition', 'value()&gt;=0'],
                ['element', 'style:apply-style-name', 'N104P0']]],
            ['tagline', 'style:style',
              ['element', 'style:name', 'Default'],
              ['element', 'style:family', 'table-cell']],
            ['tag', 'style:style',
              ['element', 'style:name', 'Result'],
              ['element', 'style:family', 'table-cell'],
              ['element', 'style:parent-style-name', 'Default'],
              ['tagline', 'style:text-properties',
                ['element', 'fo:font-style', 'italic'],
                ['element', 'style:text-underline-style', 'solid'],
                ['element', 'style:text-underline-width', 'auto'],
                ['element', 'style:text-underline-color', 'font-color'],
                ['element', 'fo:font-weight', 'bold']]],
            ['tagline', 'style:style',
              ['element', 'style:name', 'Result2'],
              ['element', 'style:family', 'table-cell'],
              ['element', 'style:parent-style-name', 'Result'],
              ['element', 'style:data-style-name', 'N104']],
            ['tag', 'style:style',
              ['element', 'style:name', 'Heading'],
              ['element', 'style:family', 'table-cell'],
              ['element', 'style:parent-style-name', 'Default'],
              ['tagline', 'style:table-cell-properties',
                ['element', 'style:text-align-source', 'fix'],
                ['element', 'style:repeat-content', 'false']],
              ['tagline', 'style:paragraph-properties',
                ['element', 'fo:text-align', 'center']],
              ['tagline', 'style:text-properties',
                ['element', 'fo:font-size', '16pt'],
                ['element', 'fo:font-style', 'italic'],
                ['element', 'fo:font-weight', 'bold']]],
            ['tag', 'style:style',
              ['element', 'style:name', 'Heading1'],
              ['element', 'style:family', 'table-cell'],
              ['element', 'style:parent-style-name', 'Heading'],
              ['tagline', 'style:table-cell-properties',
                ['element', 'style:rotation-angle', '90']]]],
          ['tag', 'office:automatic-styles',
            ['tag', 'style:page-layout',
              ['element', 'style:name', 'pm1'],
              ['tagline', 'style:page-layout-properties',
                ['element', 'style:writing-mode', 'lr-tb']],
              ['tag', 'style:header-style',
                ['tagline', 'style:header-footer-properties',
                  ['element', 'fo:min-height', '0.2957in'],
                  ['element', 'fo:margin-left', '0in'],
                  ['element', 'fo:margin-right', '0in'],
                  ['element', 'fo:margin-bottom', '0.0984in']]],
              ['tag', 'style:footer-style',
                ['tagline', 'style:header-footer-properties',
                  ['element', 'fo:min-height', '0.2957in'],
                  ['element', 'fo:margin-left', '0in'],
                  ['element', 'fo:margin-right', '0in'],
                  ['element', 'fo:margin-top', '0.0984in']]]],
            ['tag', 'style:page-layout',
              ['element', 'style:name', 'pm2'],
              ['tagline', 'style:page-layout-properties',
                ['element', 'style:writing-mode', 'lr-tb']],
              ['tag', 'style:header-style',
                ['tag', 'style:header-footer-properties',
                  ['element', 'fo:min-height', '0.2957in'],
                  ['element', 'fo:margin-left', '0in'],
                  ['element', 'fo:margin-right', '0in'],
                  ['element', 'fo:margin-bottom', '0.0984in'],
                  ['element', 'fo:border', '0.0346in solid #000000'],
                  ['element', 'fo:padding', '0.0071in'],
                  ['element', 'fo:background-color', '#c0c0c0'],
                  ['tagline', 'style:background-image']]],
              ['tag', 'style:footer-style',
                ['tag', 'style:header-footer-properties',
                  ['element', 'fo:min-height', '0.2957in'],
                  ['element', 'fo:margin-left', '0in'],
                  ['element', 'fo:margin-right', '0in'],
                  ['element', 'fo:margin-top', '0.0984in'],
                  ['element', 'fo:border', '0.0346in solid #000000'],
                  ['element', 'fo:padding', '0.0071in'],
                  ['element', 'fo:background-color', '#c0c0c0'],
                  ['tagline', 'style:background-image']]]]],
          ['tag', 'office:master-styles',
            ['tag', 'style:master-page',
              ['element', 'style:name', 'Default'],
              ['element', 'style:page-layout-name', 'pm1'],
              ['tag', 'style:header',
                ['tag', 'text:p',
                  ['data', '<text:sheet-name>???</text:sheet-name>']]],
              ['tagline', 'style:header-left',
                ['element', 'style:display', 'false']],
              ['tag', 'style:footer',
                ['tag', 'text:p',
                  ['data', 'Page <text:page-number>1</text:page-number>']]],
              ['tagline', 'style:footer-left',
                ['element', 'style:display', 'false']]],
            ['tag', 'style:master-page',
              ['element', 'style:name', 'Report'],
              ['element', 'style:page-layout-name', 'pm2'],
              ['tag', 'style:header',
                ['tag', 'style:region-left',
                  ['tag', 'text:p',
                    ['data', '<text:sheet-name>???</text:sheet-name> (<text:title>???</text:title>)']]],
                ['tag', 'style:region-right',
                  ['tag', 'text:p',
                    ['data', '<text:date style:data-style-name="N2" text:date-value="2006-09-29">09/29/2006</text:date>, <text:time>13:02:56</text:time>']]]],
              ['tagline', 'style:header-left',
                ['element', 'style:display', 'false']],
              ['tag', 'style:footer',
                ['tag', 'text:p',
                  ['data', 'Page <text:page-number>1</text:page-number> / <text:page-count>99</text:page-count>']]],
              ['tagline', 'style:footer-left',
                ['element', 'style:display', 'false']]]]]


        # Generate content.xml XML data
        xml = XML()
        self.lines = xml.convert(self.data)
        self.filedata = '\n'.join(self.lines)
        # Return generated data
        return self.filedata

class Writer:
    "Writer Class - Used to create OpenDocument Format Writer Documents."
    def __init__(self):
        "Initialize ooolib Writer instance"
        # Default to no debugging
        self.debug = False
        self.meta = Meta('odt')

    def set_meta(self, metaname, value):
        "Set meta data in your document."
        self.meta.set_meta(metaname, value)

    def save(self, filename):
        """Save .odt document

        The save function saves the current .odt document.
        """
        if self.debug: print "Writing %s" % filename
        self.savefile = zipfile.ZipFile(filename, "w")
        if self.debug: print "  meta.xml"
        self._zip_insert(self.savefile, "meta.xml", self.meta.get_meta())
        if self.debug: print "  mimetype"
        self._zip_insert(self.savefile, "mimetype", "application/vnd.oasis.opendocument.text")
        if self.debug: print "  META-INF/manifest.xml"
        self._zip_insert(self.savefile, "META-INF/manifest.xml", self._odt_manifest())
        if self.debug: print "  content.xml"
        self._zip_insert(self.savefile, "content.xml", self._odt_content())
        if self.debug: print "  settings.xml"
        # self._zip_insert(self.savefile, "settings.xml", self._odt_settings())
        if self.debug: print "  styles.xml"
        # self._zip_insert(self.savefile, "styles.xml", self._odt_styles())

    def _zip_insert(self, file, filename, data):
        now = time.localtime(time.time())[:6]
        info = zipfile.ZipInfo(filename)
        info.date_time = now
        info.compress_type = zipfile.ZIP_DEFLATED
        file.writestr(info, data)

    def _odt_manifest(self):
        "Generate odt manifest.xml data"

        self.data = ['tag', 'manifest:manifest',
          ['element', 'xmlns:manifest', 'urn:oasis:names:tc:opendocument:xmlns:manifest:1.0'],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', 'application/vnd.oasis.opendocument.text'],
            ['element', 'manifest:full-path', '/']],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', 'text/xml'],
            ['element', 'manifest:full-path', 'content.xml']],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', 'text/xml'],
            ['element', 'manifest:full-path', 'styles.xml']],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', 'text/xml'],
            ['element', 'manifest:full-path', 'meta.xml']],
          ['tagline', 'manifest:file-entry',
            ['element', 'manifest:media-type', 'text/xml'],
            ['element', 'manifest:full-path', 'settings.xml']]]

        # Generate content.xml XML data
        xml = XML()
        self.lines = xml.convert(self.data)
        self.lines.insert(1, '<!DOCTYPE manifest:manifest PUBLIC "-//OpenOffice.org//DTD Manifest 1.0//EN" "Manifest.dtd">')
        self.filedata = '\n'.join(self.lines)
        # Return generated data
        return self.filedata

    def _odt_content(self):
        "Generate odt content.xml data"

        self.data = ['tag', 'office:document-content',
          ['element', 'xmlns:office', 'urn:oasis:names:tc:opendocument:xmlns:office:1.0'],
          ['element', 'xmlns:style', 'urn:oasis:names:tc:opendocument:xmlns:style:1.0'],
          ['element', 'xmlns:text', 'urn:oasis:names:tc:opendocument:xmlns:text:1.0'],
          ['element', 'xmlns:table', 'urn:oasis:names:tc:opendocument:xmlns:table:1.0'],
          ['element', 'xmlns:draw', 'urn:oasis:names:tc:opendocument:xmlns:drawing:1.0'],
          ['element', 'xmlns:fo', 'urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0'],
          ['element', 'xmlns:xlink', 'http://www.w3.org/1999/xlink'],
          ['element', 'xmlns:dc', 'http://purl.org/dc/elements/1.1/'],
          ['element', 'xmlns:meta', 'urn:oasis:names:tc:opendocument:xmlns:meta:1.0'],
          ['element', 'xmlns:number', 'urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0'],
          ['element', 'xmlns:svg', 'urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0'],
          ['element', 'xmlns:chart', 'urn:oasis:names:tc:opendocument:xmlns:chart:1.0'],
          ['element', 'xmlns:dr3d', 'urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0'],
          ['element', 'xmlns:math', 'http://www.w3.org/1998/Math/MathML'],
          ['element', 'xmlns:form', 'urn:oasis:names:tc:opendocument:xmlns:form:1.0'],
          ['element', 'xmlns:script', 'urn:oasis:names:tc:opendocument:xmlns:script:1.0'],
          ['element', 'xmlns:ooo', 'http://openoffice.org/2004/office'],
          ['element', 'xmlns:ooow', 'http://openoffice.org/2004/writer'],
          ['element', 'xmlns:oooc', 'http://openoffice.org/2004/calc'],
          ['element', 'xmlns:dom', 'http://www.w3.org/2001/xml-events'],
          ['element', 'xmlns:xforms', 'http://www.w3.org/2002/xforms'],
          ['element', 'xmlns:xsd', 'http://www.w3.org/2001/XMLSchema'],
          ['element', 'xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance'],
          ['element', 'office:version', '1.0'],
          ['tagline', 'office:scripts'],
          ['tag', 'office:font-face-decls',
            ['tagline', 'style:font-face',
              ['element', 'style:name', 'DejaVu Sans'],
              ['element', 'svg:font-family', '&apos;DejaVu Sans&apos;'],
              ['element', 'style:font-pitch', 'variable']],
            ['tagline', 'style:font-face',
              ['element', 'style:name', 'Nimbus Roman No9 L'],
              ['element', 'svg:font-family', '&apos;Nimbus Roman No9 L&apos;'],
              ['element', 'style:font-family-generic', 'roman'],
              ['element', 'style:font-pitch', 'variable']],
            ['tagline', 'style:font-face',
              ['element', 'style:name', 'Nimbus Sans L'],
              ['element', 'svg:font-family', '&apos;Nimbus Sans L&apos;'],
              ['element', 'style:font-family-generic', 'swiss'],
              ['element', 'style:font-pitch', 'variable']]],
          ['tagline', 'office:automatic-styles'],
          ['tag', 'office:body',
            ['tag', 'office:text',
              ['tagline', 'office:forms',
                ['element', 'form:automatic-focus', 'false'],
                ['element', 'form:apply-design-mode', 'false']],
              ['tag', 'text:sequence-decls',
                ['tagline', 'text:sequence-decl',
                  ['element', 'text:display-outline-level', '0'],
                  ['element', 'text:name', 'Illustration']],
                ['tagline', 'text:sequence-decl',
                  ['element', 'text:display-outline-level', '0'],
                  ['element', 'text:name', 'Table']],
                ['tagline', 'text:sequence-decl',
                  ['element', 'text:display-outline-level', '0'],
                  ['element', 'text:name', 'Text']],
                ['tagline', 'text:sequence-decl',
                  ['element', 'text:display-outline-level', '0'],
                  ['element', 'text:name', 'Drawing']]],
              ['tagline', 'text:p',
                ['element', 'text:style-name', 'Standard']]]]]

        # Generate content.xml XML data
        xml = XML()
        self.lines = xml.convert(self.data)
        self.filedata = '\n'.join(self.lines)
        # Return generated data
        return self.filedata



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""MySQL backend"""


from cf.backends import DBConnectError
from cf.backends.dbapi2helper import DbAPI2Connection
from cf.backends.schema import *
from cf.backends import ReferenceProvider
from cf.datasources import DatasourceInfo
from cf.plugins.core import DBBackendPlugin

import time
from urllib import quote_plus

from gettext import gettext as _

import logging
log = logging.getLogger("MYSQL")

class MySQLBackend(DBBackendPlugin):

    id = "crunchyfrog.backend.mysql"
    name = _(u"MySQL Plugin")
    description = _(u"Provides access to MySQL databases")
    password_option = "password"

    def __init__(self, app):
        DBBackendPlugin.__init__(self, app)
        log.info("Activating MySQL backend")
        self.schema = MySQLSchema()
        self.reference = MySQLReferenceProvider()

    def _get_conn_opts_from_opts(self, opts):
        conn_opts = dict()
        if opts["database"]: conn_opts["db"] = opts["database"]
        if opts["host"]: conn_opts["host"] = opts["host"]
        if opts["port"]: conn_opts["port"] = opts["port"]
        if opts["user"]: conn_opts["user"] = opts["user"]
        if opts["password"]: conn_opts["passwd"] = opts["password"]
        return conn_opts

    def shutdown(self):
        log.info("Shutting down MySQL backend")

    @classmethod
    def get_datasource_options_widgets(cls, data_widgets, initial_data=None):
        return data_widgets, ["database", "host", "port", "user", "password"]

    @classmethod
    def get_label(cls, datasource_info):
        if not datasource_info.options.get("host", None):
            datasource_info.options["host"] = "localhost"
        return DBBackendPlugin.get_label(datasource_info)

    def dbconnect(self, data):
        opts = self._get_conn_opts_from_opts(data)
        if data.get("ask_for_password", False):
            pwd = self.password_prompt()
            if not pwd:
                raise DBConnectError(_(u"No password given."))
            opts["password"] = pwd
        try:
            import MySQLdb
        except ImportError:
            raise DBConnectError(_(u"Python module MySQLdb is not installed."))
        try:
            real_conn = MySQLdb.connect(**opts)
        except StandardError, err:
            raise DBConnectError(str(err))
        conn = MySQLConnection(self, self.app, real_conn)
        conn.threadsafety = MySQLdb.threadsafety
        return conn

    def test_connection(self, data):
        try:
            conn = self.dbconnect(data)
            conn.close()
        except DBConnectError, err:
            return str(err)
        return None

class MySQLConnection(DbAPI2Connection):

    def get_server_info(self):
        return "MySQL %s" % self._conn.get_server_info()

class MySQLSchema(SchemaProvider):

    def q(self, connection, sql):
        cur = connection.cursor()._cur
        cur.execute("use information_schema;")
        cur.execute(sql)
        return cur.fetchall()

    def fetch_children(self, connection, parent):
        if isinstance(parent, DatasourceInfo):
            return [SchemaCollection()]

        elif isinstance(parent, SchemaCollection):
            ret = []
            sql = "select schema_name from schemata"
            for item in self.q(connection, sql):
                ret.append(Schema(item[0]))
            return ret

        elif isinstance(parent, Schema):
            return [TableCollection(schema=parent),
                    ViewCollection(schema=parent)]

        elif isinstance(parent, TableCollection) \
        or isinstance(parent, ViewCollection):
            if isinstance(parent, TableCollection):
                obj = Table
                kind = "BASE TABLE"
            elif isinstance(parent, ViewCollection):
                obj = View
                kind = "VIEW"
            ret = []
            schema = parent.get_data("schema")
            sql = "select table_name, table_comment from tables \
            where table_schema = '%(schema)s' and table_type = '%(kind)s'" % {"schema" : schema.name, "kind" : kind}
            for item in self.q(connection, sql):
                ret.append(obj(item[0], item[1], schema=schema))
            return ret

        elif isinstance(parent, Table):
            return [ColumnCollection(table=parent)]

        elif isinstance(parent, View):
            return [ColumnCollection(table=parent)]

        elif isinstance(parent, ColumnCollection):
            ret = []
            table = parent.get_data("table")
            sql = "select column_name, column_comment from columns \
            where table_name = '%(table)s' and table_schema = '%(schema)s'" % {"table" : table.name, "schema": table.get_data("schema").name}
            for item in self.q(connection, sql):
                ret.append(Column(item[0], item[1], table=table))
            return ret

class MySQLReferenceProvider(ReferenceProvider):
    name = _(u"MySQL Reference")
    base_url = "http://dev.mysql.com/doc/refman/5.1/en/index.html"

    def get_context_help_url(self, term):
        url = "http://dev.mysql.com/doc/mysql/search.php?version=5.1&q="
        url += quote_plus(term.strip())
        return url


try:
    import MySQLdb
except ImportError:
    MySQLBackend.INIT_ERROR = _(u"Python module MySQLdb required.")



# Pretty print SQL statements
# by Peter Bengtsson, www.peterbe.com
# August, 2004

# Some modifications by Andi Albrecht, March 2008

import re

__version__='0.1'

def printsql(sql):
    """ simple version of pprint for SQL strings """
    keywords = ('select','from','where','order by','desc','asc',
                'limit','offset','update','delete','insert','set',
                'having','group by','count','table','create',
                'drop','and','or','in','ilike','like',
                'inner join', 'outer join', 'left outer join',
                'join'
                )
    aloner = lambda x: '(\s%s\s|^%s\s|\s%s$)'%(x,x,x)

    re_flags = re.I|re.MULTILINE
    regex = '|'.join(map(aloner, keywords))
    anykeyword = re.compile(regex, re_flags)
    for each in anykeyword.findall(sql):
        each = list(each)
        while '' in each:each.remove('')
        for subeach in each:
            sql = sql.replace(subeach, subeach.upper())

    if len(sql) > 10:

        spad = '  '

        regex = re.compile('(SELECT\s(.*?))(\sFROM\s(.*?)\s)', re_flags)
        found = regex.findall(sql)
        if found:
            sql = sql.replace(found[0][0], 'SELECT\n%s%s\n'%(spad, found[0][1]))
            sql = sql.replace(found[0][2], 'FROM\n %s\n'%found[0][3])

        regex = re.compile('(ORDER BY\s(.*?))(\sLIMIT\s(.*?)\s)', re_flags)
        found = regex.findall(sql)
        if found:
            sql = sql.replace(found[0][0], '\nORDER BY %s\n'%found[0][1])
            sql = sql.replace(found[0][2], 'LIMIT %s\n'%found[0][3])

        regex = re.compile('(\s+WHERE\s(.*?)\s(GROUP BY|ORDER BY|LIMIT))', re_flags)
        found = regex.findall(sql)
        if found:
            sql = sql.replace(found[0][0],
                              '\nWHERE\n%s%s\n%s'%(spad, found[0][1],found[0][2]))

    return sql


def test(s):
    print printsql(s)
    print

if __name__=='__main__':
    print "TESTING SOME SQL"
    print 
    test('select * from foo order by bar;')
    test('''select id, bull, time from foobartable
    where time=1 and shit='yes' order by bar limit 30 offset 10;''')



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""DBus service and client"""

import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop

main_loop = DBusGMainLoop()
bus = dbus.SessionBus(mainloop=main_loop)

DBUS_NAME = "org.gnome.crunchyfrog"
DBUS_PATH = "/org/gnome/crunchyfrog"

class CFService(dbus.service.Object):
    """DBus service"""

    def __init__(self, app):
        global bus
        bus_name = dbus.service.BusName(DBUS_NAME, bus=bus)
        dbus.service.Object.__init__(self, bus_name, DBUS_PATH)
        self.app = app

    @dbus.service.method(DBUS_NAME)
    def new_instance(self):
        """Creates a new instance.

        :Returns: Instance ID
        """
        instance = self.app.new_instance()
        return instance.get_data("instance-id")

    @dbus.service.method(DBUS_NAME)
    def open_uri(self, instance_id, uri):
        """Opens an URI

        :Parameter:
            instance_id
                An instance ID
            uri
                An URI
        """
        instance = self.app.get_instance(instance_id)
        editor = instance.new_editor(uri)

    @dbus.service.method(DBUS_NAME)
    def get_instances(self):
        """Returns running instances.

        The return value is a list of 2-tuples instance_id, window title.
        """
        ret = []
        for key, value in self.app.get_data("instances").items():
            ret.append((key, value.widget.get_title()))
        return ret


def get_client():
    """Returns a DBus client"""
    global bus
    return bus.get_object(DBUS_NAME, DBUS_PATH)



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""SQL editor and results view"""

from gettext import gettext as _
import logging
import os
import re
import string
import thread
import time

import gconf
import gnome
import gnomevfs
import gobject
import gtk
import pango

from kiwi.ui import dialogs

from cf import sqlparse
from cf.backends import Query
from cf.ui import GladeWidget
from cf.ui.confirmsave import ConfirmSaveDialog
from cf.ui.toolbar import CFToolbar
from cf.ui.widgets import DataExportDialog
from cf.ui.widgets.grid import Grid
from cf.ui.widgets.sqlview import SQLView


class Editor(GladeWidget):

    __gsignals__ = {
        "connection-changed" : (gobject.SIGNAL_RUN_LAST,
                                gobject.TYPE_NONE,
                                (gobject.TYPE_PYOBJECT,))
    }

    def __init__(self, app, instance):
        self.app = app
        self.instance = instance
        self.connection = None
        self.__conn_close_tag = None
        self._query_timer = None
        self._filename = None
        self._filecontent_read = ""
        GladeWidget.__init__(self, app, "crunchyfrog", "box_editor")
        self.set_data("win", None)
        self.show_all()

    # Widget setup

    def _setup_widget(self):
        self._setup_textview()
        self._setup_resultsgrid()

    def _setup_textview(self):
        self.textview = SQLView(self.app, self)
        sw = self.xml.get_widget("sw_editor_textview")
        sw.add(self.textview)

    def _setup_resultsgrid(self):
        self.results = ResultsView(self.app, self.instance, self.xml)

    def _setup_connections(self):
        self.textview.connect("populate-popup", self.on_populate_popup)

    # Callbacks

    def on_close(self, *args):
        self.close()

    def on_connection_closed(self, connection):
        if connection == self.connection and self.__conn_close_tag:
            connection.disconnect(self.__conn_close_tag)
        self.set_connection(None)

    def on_explain(self, *args):
        gobject.idle_add(self.explain)

    def on_populate_popup(self, textview, popup):
        cfg = self.app.config
        sep = gtk.SeparatorMenuItem()
        sep.show()
        popup.append(sep)
        if self.connection and self.connection.provider.reference:
            refviewer = self.instance.get_data("refviewer")
            buffer = self.textview.get_buffer()
            bounds = buffer.get_selection_bounds()
            if bounds:
                url = self.connection.provider.reference.get_context_help_url(buffer.get_text(*bounds))
            else:
                url = None
            if url and refviewer:
                item = gtk.ImageMenuItem("gtk-help")
                item.connect("activate", self.on_show_context_help,
                             refviewer, url)
                item.show()
                popup.append(item)
        item = gtk.CheckMenuItem(_(u"Split statements"))
        item.set_active(cfg.get("sqlparse.enabled"))
        item.connect("toggled", lambda x: cfg.set("sqlparse.enabled",
                                                  x.get_active()))
        item.show()
        popup.append(item)
        if self.get_data("win"):
            lbl = _(u"Show in main window")
            cb = self.on_show_in_main_window
        else:
            lbl = _(u"Show in separate window")
            cb = self.on_show_in_separate_window
        item = gtk.MenuItem(lbl)
        item.connect("activate", cb)
        item.show()
        popup.append(item)
        item = gtk.ImageMenuItem("gtk-close")
        item.show()
        item.connect("activate", self.on_close)
        popup.append(item)

    def on_query_started(self, query):
        start = time.time()
        # Note: The higher the time out value, the longer the query takes.
        #    50 is a reasonable value anyway.
        #    The start time is for the UI only. The real execution time
        #    is calculated in the Query class.
        if self._query_timer is not None:
            gobject.source_remove(self._query_timer)
        self.results.add_separator()
        self.results.add_message(query.statement, type_='query')
        query.path_status = self.results.add_message("")
        self._query_timer = gobject.timeout_add(50, self.update_exectime,
                                                start, query)

    def on_query_finished(self, query, tag_notice):
        if self._query_timer:
            gobject.source_remove(self._query_timer)
            self._query_timer = None
        self.results.set_query(query)
        if query.failed:
            msg = _(u'Query failed (%.3f seconds)') % query.execution_time
            type_ = 'error'
        elif query.description:
            msg = (_(u"Query finished (%.3f seconds, %s rows)")
                   % (query.execution_time, query.rowcount))
            type_ = 'info'
        else:
            msg = _(u"Query finished (%.3f seconds, %s affected rows)")
            msg = msg % (query.execution_time, query.rowcount)
            type_ = 'info'
        self.results.add_message(msg, type_, query.path_status)
        if self.connection.handler_is_connected(tag_notice):
            self.connection.disconnect(tag_notice)
        self.textview.grab_focus()

    def on_show_context_help(self, menuitem, refviewer, url):
        refviewer.load_url(url)

    def on_show_in_main_window(self, *args):
        gobject.idle_add(self.show_in_main_window)

    def on_show_in_separate_window(self, *args):
        gobject.idle_add(self.show_in_separate_window)

    # Public methods

    def close(self):
        """Close editor, displays a confirmation dialog for unsaved files."""
        dlg = ConfirmSaveDialog(self, [self])
        resp = dlg.run()
        if resp == 1:
            ret = dlg.save_files()
        elif resp == 2:
            ret = True
        else:
            ret = False
        dlg.destroy()
        if ret:
            if self.get_data("win"):
                self.get_data("win").destroy()
            else:
                self.destroy()
            self.instance._editors.remove(self)
            return True

    def commit(self):
        """Commit current transaction, if any."""
        if not self.connection: return
        cur = self.connection.cursor()
        cur.execute("commit")
        cur.close()

    def rollback(self):
        """Commit current transaction, if any."""
        if not self.connection: return
        cur = self.connection.cursor()
        cur.execute("rollback")
        cur.close()

    def begin_transaction(self):
        """Begin transaction."""
        if not self.connection: return
        cur = self.connection.cursor()
        cur.execute("begin")
        cur.close()

    def execute_query(self):
        def exec_threaded(statement):
            cur = self.connection.cursor()
            if self.app.config.get("sqlparse.enabled", True):
                if self.connection:
                    dialect = self.connection.sqlparse_dialect
                else:
                    dialect = None
                stmts = sqlparse.sqlsplit(statement, dialect=dialect)
            else:
                stmts = [statement]
            for stmt in stmts:
                if not stmt.strip():
                    continue
                query = Query(stmt, cur)
                query.coding_hint = self.connection.coding_hint
                gtk.gdk.threads_enter()
                query.connect("started", self.on_query_started)
                query.connect("finished",
                              self.on_query_finished,
                              tag_notice)
                gtk.gdk.threads_leave()
                query.execute(True)
                if query.failed:
                    # hmpf, doesn't work that way... so just return here...
                    return
#                    gtk.gdk.threads_enter()
#                    dlg = gtk.MessageDialog(None,
#                                            gtk.DIALOG_MODAL|
#                                            gtk.DIALOG_DESTROY_WITH_PARENT,
#                                            gtk.MESSAGE_ERROR,
#                                            gtk.BUTTONS_YES_NO,
#                                            _(u"An error occurred. Continue?"))
#                    if dlg.run() == gtk.RESPONSE_NO:
#                        leave = True
#                    else:
#                        leave = False
#                    dlg.destroy()
#                    gtk.gdk.threads_leave()
#                    if leave:
#                        return
        buffer = self.textview.get_buffer()
        bounds = buffer.get_selection_bounds()
        self.results.reset()
        if not bounds:
            bounds = buffer.get_bounds()
        statement = buffer.get_text(*bounds)
        if self.app.config.get("editor.replace_variables"):
            tpl = string.Template(statement)
            tpl_search = tpl.pattern.search(tpl.template)
            if tpl_search and tpl_search.groupdict().get("named"):
                dlg = StatementVariablesDialog(tpl)
                gtk.gdk.threads_enter()
                if dlg.run() == gtk.RESPONSE_OK:
                    statement = dlg.get_statement()
                else:
                    statement = None
                dlg.destroy()
                gtk.gdk.threads_leave()
                if not statement:
                    return
        def foo(connection, msg):
            self.results.add_message(msg)
        tag_notice = self.connection.connect("notice", foo)
        if self.connection.threadsafety >= 2:
            thread.start_new_thread(exec_threaded, (statement,))
        else:
            cur = self.connection.cursor()
            if self.app.config.get("sqlparse.enabled", True):
                if self.connection:
                    dialect = self.connection.sqlparse_dialect
                else:
                    dialect = None
                stmts = sqlparse.sqlsplit(statement, dialect=dialect)
            else:
                stms = [statement]
            for stmt in stmts:
                if not stmt.strip():
                    continue
                query = Query(stmt, cur)
                query.coding_hint = self.connection.coding_hint
                query.connect("started", self.on_query_started)
                query.connect("finished", self.on_query_finished, tag_notice)
                query.execute()

    def explain(self):
        buffer = self.textview.get_buffer()
        bounds = buffer.get_selection_bounds()
        if not bounds:
            bounds = buffer.get_bounds()
        statement = buffer.get_text(*bounds)
        if self.connection:
            dialect = self.connection.sqlparse_dialect
        else:
            dialect = None
        if len(sqlparse.sqlsplit(statement, dialect=dialect)) > 1:
            dialogs.error(_(u"Select a single statement to explain."))
            return
        data = []
        if self.connection:
            data = self.connection.explain(statement)
        self.results.set_explain(data)

    def set_connection(self, conn):
        if self.connection and self.__conn_close_tag:
            self.connection.disconnect(self.__conn_close_tag)
            self.__conn_close_tag = None
        self.connection = conn
        if conn:
            self.__conn_close_tag = self.connection.connect("closed",
                                                            self.on_connection_closed)
        else:
            self._conn_close_tag = None
        self.emit("connection-changed", conn)

    def set_filename(self, filename):
        self._filename = filename
        if filename:
            f = open(self._filename)
            a = f.read()
            f.close()
        else:
            a = ""
        self._filecontent_read = a
        self.textview.get_buffer().set_text(a)
        self.app.recent_manager.add_item(gnomevfs.get_uri_from_local_path(filename))

    def get_filename(self):
        return self._filename

    def file_contents_changed(self):
        if self._filename:
            buffer = self.textview.get_buffer()
            return buffer.get_text(*buffer.get_bounds()) != self._filecontent_read
        return False

    def contents_changed(self):
        if self._filename:
            return self.file_contents_changed()
        elif len(self.get_text()) == 0:
            return False
        return True

    def file_confirm_save(self):
        dlg = dialogs.yesno(_(u"Save file %(name)s before closing the editor?") % {"name":os.path.basename(self._filename)})
        if dlg == gtk.RESPONSE_YES:
            return self.save_file_as()
        return True

    def save_file(self, parent=None, default_name=None):
        if not self._filename:
            return self.save_file_as(parent=parent, default_name=default_name)
        buffer = self.get_buffer()
        a = buffer.get_text(*buffer.get_bounds())
        f = open(self._filename, "w")
        f.write(a)
        f.close()
        self._filecontent_read = a
        gobject.idle_add(buffer.emit, "changed")
        return True

    def save_file_as(self, parent=None, default_name=None):
        if not parent:
            parent = self.instance.widget
        dlg = gtk.FileChooserDialog(_(u"Save file"),
                            parent,
                            gtk.FILE_CHOOSER_ACTION_SAVE,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        if self._filename:
            dlg.set_filename(self._filename)
        else:
            dlg.set_current_folder(self.app.config.get("editor.recent_folder", ""))
            if default_name:
                dlg.set_current_name(default_name)
        filter = gtk.FileFilter()
        filter.set_name(_(u"All files (*)"))
        filter.add_pattern("*")
        dlg.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name(_(u"SQL files (*.sql)"))
        filter.add_pattern("*.sql")
        dlg.add_filter(filter)
        dlg.set_filter(filter)
        if dlg.run() == gtk.RESPONSE_OK:
            self._filename = dlg.get_filename()
            gobject.idle_add(self.save_file)
            self.app.config.set("editor.recent_folder", dlg.get_current_folder())
            ret = True
        else:
            ret = False
        dlg.destroy()
        return ret

    def get_buffer(self):
        return self.textview.get_buffer()

    def get_text(self):
        buffer = self.get_buffer()
        return buffer.get_text(*buffer.get_bounds())

    def set_text(self, txt):
        buffer = self.get_buffer().set_text(txt)

    def show_in_separate_window(self):
        win = EditorWindow(self.app, self.instance,
                           editor=self)
        win.show_all()
        self.set_data("win", win)

    def show_in_main_window(self):
        self.instance.queries.attach(self)
        win = self.get_data("win")
        if win:
            win.destroy()
        self.set_data("win", None)

    def update_exectime(self, start, query):
        lbl = _("Query running... (%.3f seconds)" % (time.time()-start))
        self.results.add_message(lbl, path=query.path_status)
        if query.executed:
            if self._query_timer is not None:
                gobject.source_remove(self._query_timer)
            self._query_timer = None
            return False
        else:
            return True


class EditorWindow(GladeWidget):

    def __init__(self, app, instance, editor=None):
        GladeWidget.__init__(self, app, "crunchyfrog", "editorwindow")
        self.instance = instance
        self.editor = None
        self.toolbar = CFToolbar(self.instance, xml="crunchyfrog",
                                 cb_provider=self)
        box = self.xml.get_widget("mainbox_editor")
        box.pack_start(self.toolbar.widget, False, False)
        box.reorder_child(self.toolbar.widget, 1)
        item = gtk.ToolButton()
        item.set_label(_(u"Show in main window"))
        item.set_stock_id("gtk-dnd-multiple")
        item.connect("clicked", self.on_show_in_main_window)
        item.show()
        self.toolbar.insert(item, -1)
        item = gtk.ToolButton()
        item.set_stock_id("gtk-close")
        item.connect("clicked", self.on_close)
        item.show()
        self.toolbar.insert(item, -1)
        self.restore_window_state()
        if editor is not None:
            self.attach(editor)

    def on_buffer_changed(self, buffer):
        gobject.idle_add(self.update_title)

    def on_close(self, *args):
        self.close()

    def on_execute_query(self, *args):
        gobject.idle_add(self.editor.execute_query)

    def on_begin_transaction(self, *args):
        gobject.idle_add(self.editor.begin_transaction)

    def on_commit(self, *args):
        gobject.idle_add(self.editor.commit)

    def on_rollback(self, *args):
        gobject.idle_add(self.editor.rollback)

    def on_query_new(self, *args):
        self.instance.on_query_new(self, *args)

    def on_open_file(self, *args):
        dlg = gtk.FileChooserDialog(_(u"Select file"),
                            self.widget,
                            gtk.FILE_CHOOSER_ACTION_OPEN,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dlg.set_current_folder(self.app.config.get("editor.recent_folder", ""))
        filter = gtk.FileFilter()
        filter.set_name(_(u"All files (*)"))
        filter.add_pattern("*")
        dlg.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name(_(u"SQL files (*.sql)"))
        filter.add_pattern("*.sql")
        dlg.add_filter(filter)
        dlg.set_filter(filter)
        if dlg.run() == gtk.RESPONSE_OK:
            gobject.idle_add(self.editor.set_filename, dlg.get_filename())
            self.app.config.set("editor.recent_folder", dlg.get_current_folder())
        dlg.destroy()

    def on_save_file(self, *args):
        self.editor.save_file(parent=self.widget)

    def on_save_file_as(self, *args):
        self.editor.save_file_as(parent=self.widget)

    def on_copy(self, *args):
        self.editor.textview.get_buffer().copy_clipboard(gtk.clipboard_get())

    def on_paste(self, *args):
        self.editor.textview.get_buffer().paste_clipboard(gtk.clipboard_get(), None, True)

    def on_cut(self, *args):
        self.editor.textview.get_buffer().cut_clipboard(gtk.clipboard_get(), True)

    def on_delete(self, *args):
        self.editor.textview.get_buffer().delete_selection(True, True)

    def on_show_in_main_window(self, *args):
        self.editor.show_in_main_window()

    def on_configure_event(self, win, event):
        config = self.app.config
        if not config.get("querywin.maximized"):
            config.set("querywin.width", event.width)
            config.set("querywin.height", event.height)

    def on_window_state_event(self, win, event):
        config = self.app.config
        bit = gtk.gdk.WINDOW_STATE_MAXIMIZED.value_names[0] in event.new_window_state.value_names
        config.set("querywin.maximized", bit)

    def attach(self, editor):
        self.editor = editor
        box = self.xml.get_widget("mainbox_editor")
        ebox = self.xml.get_widget("box_editor")
        ebox.get_parent().remove(ebox)
        if self.editor.get_parent():
            self.editor.reparent(box)
            expand, fill, padding, pack_type = box.query_child_packing(self.editor.widget)
            box.set_child_packing(self.editor.widget, True, True, padding, gtk.PACK_START)
        else:
            box.pack_start(self.editor.widget, True, True)
        box.reorder_child(self.editor.widget, 2)
        buffer = self.editor.textview.get_buffer()
        buffer.connect("changed", self.on_buffer_changed)
        self.toolbar.set_editor(editor)
        self.update_title()

    def close(self):
        self.editor.close()

    def restore_window_state(self):
        if self.app.config.get("querywin.width", -1) != -1:
            self.widget.resize(self.app.config.get("querywin.width"),
                               self.app.config.get("querywin.height"))
        if self.app.config.get("querywin.maximized", False):
            self.widget.maximize()

    def update_title(self):
        buffer = self.editor.textview.get_buffer()
        txt = buffer.get_text(*buffer.get_bounds())
        txt = re.sub("\s+", " ", txt)
        txt = txt.strip()
        if len(txt) > 28:
            txt = txt[:25]+"..."
        if txt:
            txt = _(u"Query") + ": "+txt
        else:
            txt = _(u"Query")
        self.set_title(txt)


class ResultsView(GladeWidget):

    def __init__(self, app, instance, xml):
        self.instance = instance
        GladeWidget.__init__(self, app, xml, "editor_results")

    def _setup_widget(self):
        self.grid = ResultList(self.app, self.instance, self.xml)
        self.messages = self.xml.get_widget("editor_results_messages")
        model = gtk.ListStore(str,  # 0 stock id
                              str,  # 1 message
                              str,  # 2 foreground color
                              int,  # 3 font weight
                              bool, # 4 is separator row
                              str,  # 5 font description
                              )
        model.connect("row-inserted", self._msg_model_changed)
        model.connect("row-deleted", self._msg_model_changed)
        self.messages.set_model(model)
        col = gtk.TreeViewColumn()
        renderer = gtk.CellRendererPixbuf()
        col.pack_start(renderer, expand=False)
        col.add_attribute(renderer, 'stock-id', 0)
        renderer = gtk.CellRendererText()
        col.pack_start(renderer, expand=True)
        col.add_attribute(renderer, 'text', 1)
        col.add_attribute(renderer, 'foreground', 2)
        col.add_attribute(renderer, 'weight', 3)
        col.add_attribute(renderer, 'font', 5)
        self.messages.append_column(col)
        self.messages.set_row_separator_func(self._set_row_separator)
        self.explain_model = gtk.ListStore(str)
        treeview = self.xml.get_widget("editor_explain")
        treeview.set_model(self.explain_model)
        col = gtk.TreeViewColumn("", gtk.CellRendererText(), text=0)
        treeview.append_column(col)

    def _msg_model_changed(self, model, *args):
        tb_clear = self.xml.get_widget("tb_messages_clear")
        tb_copy = self.xml.get_widget("tb_messages_copy")
        sensitive = model.get_iter_first() is not None
        tb_clear.set_sensitive(sensitive)
        tb_copy.set_sensitive(sensitive)

    def _setup_connections(self):
        self.grid.grid.connect("selection-changed",
                               self.on_grid_selection_changed)

    def _set_row_separator(self, model, iter):
        return model.get_value(iter, 4)

    def on_copy_data(self, *args):
        gobject.idle_add(self.copy_data)

    def on_export_data(self, *args):
        gobject.idle_add(self.export_data)

    def on_messages_clear(self, *args):
        self.messages.get_model().clear()

    def on_messages_copy(self, *args):
        model = self.messages.get_model()
        iter_ = model.get_iter_first()
        plain = []
        while iter_ is not None:
            if model.get_value(iter_, 4): # Is it a separator?
                plain.append('-'*20)
            else:
                value = model.get_value(iter_, 1)
                if value is not None:
                    plain.append(value)
            iter_ = model.iter_next(iter_)
        clipboard = gtk.clipboard_get()
        clipboard.set_text('\n'.join(plain))

    def on_grid_selection_changed(self, grid, selected_cells):
        self.xml.get_widget("editor_copy_data").set_sensitive(bool(selected_cells))

    def copy_data(self):
        rows = dict()
        for cell in self.grid.grid.get_selected_cells():
            value = self.grid.grid.get_cell_data(cell, repr=True)
            if rows.has_key(cell[0]):
                rows[cell[0]] += "%s\t" % value
            else:
                rows[cell[0]] = "%s\t" % value
        rownums = rows.keys()
        rownums.sort()
        txt = ""
        for rownum in rownums:
            txt += "%s\n" % rows.get(rownum)
        display = gtk.gdk.display_manager_get().get_default_display()
        clipboard = gtk.Clipboard(display, "CLIPBOARD")
        clipboard.set_text(txt)
        self.instance.statusbar.set_message(_(u"Data copied to clipboard"))

    def export_data(self):
        data = self.grid.grid.get_grid_data()
        description = self.grid.grid.description
        selected = self.grid.grid.get_selected_rows()
        statement = self.grid.query.statement
        gtk.gdk.threads_enter()
        dlg = DataExportDialog(self.app, self.instance.widget,
                               data, selected, statement, description)
        if dlg.run() == gtk.RESPONSE_OK:
            dlg.hide()
            dlg.export_data()
        dlg.destroy()
        gtk.gdk.threads_leave()

    def reset(self):
        # Explain
        self.explain_model.clear()
        # Messages
        self.messages.get_model().clear()
        self.set_current_page(2)

    def set_explain(self, data):
        self.explain_model.clear()
        for item in data:
            iter = self.explain_model.append()
            self.explain_model.set(iter, 0, item)

    def set_query(self, query):
        self.grid.set_query(query)
        model = self.messages.get_model()
        for err in query.errors:
            self.add_error(err.strip(), monospaced=True)
        for msg in query.messages:
            self.add_output(msg.strip())
        if query.errors:
            curr_page = 2
        elif query.description:
            curr_page = 0
        else:
            curr_page = 2
        self.xml.get_widget("editor_export_data").set_sensitive(bool(query.rows))
        gobject.idle_add(self.set_current_page, curr_page)

    def add_message(self, msg, type_=None, path=None, monospaced=False):
        """Add a message.

        Args:
          msg: The message to add.
          type_: Message type ('info', 'output',
                 'error', 'warning', 'query', None).
        """
        assert type_ in (None, 'info', 'output', 'error', 'warning', 'query')
        stock_id = None
        foreground = None
        if monospaced:
            font = 'Monospace 10'
        else:
            font = None
        weight = pango.WEIGHT_NORMAL
        if type_ == 'info':
            stock_id = 'gtk-info'
            foreground = '#336699'
        elif type_ == 'error':
            stock_id = 'gtk-dialog-error'
            foreground = '#a40000'
            weight = pango.WEIGHT_BOLD
        elif type_ == 'warning':
            stock_id = 'gtk-dialog-warning'
            foreground = '#00a400'
        elif type_ == 'query':
            stock_id ='gtk-execute'
            weight = pango.WEIGHT_BOLD
        elif type_ == 'output':
            stock_id = 'gtk-go-back'
            font = 'Monospace'
        model = self.messages.get_model()
        if model is None:  # we are in shutdown phase
            return
        msg = msg.strip()
        if path is None:
            itr = model.append([stock_id, msg, foreground, weight,
                                False, font])
        else:
            try:
                iter_ = model.get_iter(path)
            except ValueError:
                return
            model.set_value(iter_, 0, stock_id)
            model.set_value(iter_, 1, msg)
            model.set_value(iter_, 2, foreground)
            model.set_value(iter_, 3, weight)
            model.set_value(iter_, 5, font)
            itr = iter_
        self.messages.scroll_to_cell(model.get_path(itr))
        return model.get_path(itr)

    def add_error(self, msg, monospaced=False):
        return self.add_message(msg, 'error', monospaced=monospaced)

    def add_info(self, msg):
        return self.add_message(msg, 'info')

    def add_warning(self, msg):
        return self.add_message(msg, 'warning')

    def add_output(self, msg):
        return self.add_message(msg, 'output')

    def add_separator(self):
        model = self.messages.get_model()
        if not model.get_iter_first():
            return
        model.append([None, None, None, pango.WEIGHT_NORMAL, True, None])


class ResultList(GladeWidget):
    """Result list with toolbar"""

    def __init__(self, app, instance, xml):
        GladeWidget.__init__(self, app, xml, "editor_results_data")
        self.instance = instance
        self.query = None

    def _setup_widget(self):
        self.grid = Grid()
        self.xml.get_widget("sw_grid").add(self.grid)

    def set_query(self, query):
        self.query = query
        self.grid.reset()
        if self.query.description:
            self.grid.set_result(self.query.rows, self.query.description,
                                 self.query.coding_hint)



class StatementVariablesDialog(gtk.Dialog):

    def __init__(self, template):
        gtk.Dialog.__init__(self, _(u"Variables"),
                            None,
                            gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_OK, gtk.RESPONSE_OK))
        self.template = template
        self._widgets = dict()
        self._setup_widget()

    def _setup_widget(self):
        vars = []
        found = []
        for match in self.template.pattern.finditer(self.template.template):
            name = match.groupdict().get("named")
            if not name or name.lower() in found:
                continue
            vars.append(name)
            found.append(name.lower())
        table = gtk.Table(len(vars), 2)
        table.set_row_spacings(5)
        table.set_col_spacings(7)
        for i in range(len(vars)):
            lbl = gtk.Label(vars[i])
            lbl.set_alignment(0, 0.5)
            table.attach(lbl, 0, 1, i, i+1, gtk.FILL, gtk.FILL)
            entry = gtk.Entry()
            table.attach(entry, 1, 2, i, i+1, gtk.EXPAND|gtk.FILL, gtk.FILL)
            self._widgets[vars[i]] = entry
        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        sw.add_with_viewport(table)
        sw.set_border_width(10)
        self.vbox.pack_start(sw, True, True)
        self.vbox.show_all()

    def on_value_edited(self, renderer, path, value):
        model = self.treeview.get_model()
        iter = model.get_iter(path)
        model.set_value(iter, 1, value)

    def get_statement(self):
        data = dict()
        for var, widget in self._widgets.items():
            data[var] = widget.get_text()
        return self.template.safe_substitute(data)



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""Helper classes for DB-API2 connections"""

import gtk

from kiwi.ui import dialogs

import sys

from gettext import gettext as _

from cf.backends import DBConnection, DBCursor

class DbAPI2Cursor(DBCursor):

    def __init__(self, connection):
        DBCursor.__init__(self, connection)
        self._cur = self.connection._conn.cursor()

    def _get_description(self):
        return self._cur.description
    description = property(fget=_get_description)

    def _get_rowcount(self):
        return self._cur.rowcount
    rowcount = property(fget=_get_rowcount)

    def execute(self, statement):
        try:
            self._cur.execute(statement)
            exc = None
        except:
            exc = sys.exc_info()
        self.connection.update_transaction_status()
        if exc:
            raise exc[1]

    def fetchall(self):
        return self._cur.fetchall()

    def close(self):
        self._cur.close()

class DbAPI2Connection(DBConnection):

    cursor_class = DbAPI2Cursor

    def __init__(self, provider, app, real_conn):
        DBConnection.__init__(self, provider, app)
        self._conn = real_conn

    def close(self):
        self._conn.close()
        DBConnection.close(self)

    def cursor(self):
        return self.cursor_class(self)

    def commit(self):
        self._conn.commit()
        self.update_transaction_status()

    def rollback(self):
        self._conn.rollback()
        self.update_transaction_status()

    def explain(self, statement):
        try:
            sql = "EXPLAIN %s" % statement
            cur = self._conn.cursor()
            cur.execute(sql)
            data = cur.fetchall()
            cur.close()
        except:
            gtk.gdk.threads_enter()
            dialogs.error(_(u"An error occured"), str(sys.exc_info()[1]))
            gtk.gdk.threads_leave()
            return []
        return data



# Copyright (C) 2008 Andi Albrecht, albrecht.andi@gmail.com
#
# This module is part of python-sqlparse and is released under
# the BSD License: http://www.opensource.org/licenses/bsd-license.php.

"""Statement tokenizer."""

import logging
import re


def _tokenize(statement, dialect):
    """Tokenize a SQL statement.

    Args:
        statement: SQL statement as string.
        dialect: A Dialect instance.

    Returns:
        List of tokens.
    """
    prev = None
    tokens = []
    for match in dialect.tokenizer_regex.finditer(statement):
        if match.start() != match.end():
            token =Token(statement[match.start():match.end()], dialect)
            if prev:
                prev._next = token
            token._prev = prev
            tokens.append(token)
            prev = token
    return tokens


def _join_tokens(tokens):
    """Reverse _tokenize()."""
    return ''.join(str(x) for x in tokens)


def _split_statements(tokens, dialect):
    """Split list of tokens into statements.

    Args:
        tokens: Tokens
        dialect: A Dialect instance.

    Returns:
        tuple where each element is a list of tokens for a statement.
    """
    statements = []
    statement = []
    statement_finished = False
    semicolon_split_level = 0
    for token in tokens:
        statement.append(token)
        semicolon_split_level += dialect.semicolon_split_level(token.upper())
        if token.type == token.SEMICOLON and not semicolon_split_level:
            statement_finished = True
            continue
        if statement_finished:
            new_token = statement.pop()
            statements.append(statement)
            statement = [new_token]
            statement_finished = False
    if statement:
        statements.append(statement)
    statements = tuple(statements)
    logging.debug('_split_statements() --> %r', statements)
    return statements


def _is_multistatement(tokens, dialect):
    """Returns True if tokens contains more than one statement.

    Args:
        tokens: Token list.
        dialect: A Dialect instance.
    """
    return len(_split_statements(tokens, dialect)) > 1


class Token(object):
    """Represents a token."""

    UNKNOWN = 0
    SEMICOLON = 1
    WHITESPACE = 2
    OPERATOR = 3

    def __init__(self, raw, dialect):
        """Constructor.

        Args:
          raw: The 'raw' token.
          dialect: The dialect used.
        """
        self._next = None
        self._prev = None
        self._raw = raw
        self._dialect = dialect

    def __repr__(self):
        if len(self._raw) > 7:
            short = self.upper()[:6]+'...'
        else:
            short = self.upper()
        return '<Token \'%s\' at 0x%07x>' % (short, id(self))

    def __str__(self):
        return str(self._raw)

    def __unicode__(self):
        return unicode(self._raw)

    def __cmp__(self, other):
        if isinstance(other, basestring):
            return cmp(self.upper(), other.upper())
        return cmp(self, other)

    @property
    def raw(self):
        return self._raw

    @property
    def type(self):
        if self._raw == ';':
            return self.SEMICOLON
        elif re.compile(r'\s+$').match(self._raw):
            return self.WHITESPACE
        elif self.upper() in self._dialect.operators:
            return self.OPERATOR
        else:
            return self.UNKNOWN

    def upper(self):
        return self._raw.upper()




# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2007 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""User database

The user database can be used by plugins to store some information.
It is also used by CrunchyFrog itself i.e. to store connection data.

:USER_DB: Path to sqlite3 file
"""

import gobject

import os.path
try:
    import sqlite3 as sqlite
except ImportError:
    from pysqlite2 import dbapi2 as sqlite

from cf import USER_DIR

USER_DB = os.path.join(USER_DIR, "user.db")

import logging
log = logging.getLogger("USERDB")

class UserDB(gobject.GObject):
    """User database class.

    An instance of this class is accessible through the ``userdb``
    attribute of an `CFApplication`_ instance.

    Usage example
    =============

        .. sourcecode:: python

            >>> app.userdb.get_table_version("foo")
            None
            >>> app.userdb.create_table("foo", "0.1", create_statement)
            >>> app.userdb.get_table_version("foo")
            '0.1'
            >>> app.userdb.cursor.execute("insert into foo (value) values (?)", ("bar",))
            >>> app.userdb.conn.commit()
            >>> app.userdb.cursor.execute("select * from foo")
            >>> app.userdb.cursor.fetchone()
            ('bar',)
            >>> app.userdb.drop_table("foo")


    Instance attributes
    ===================

        :conn: Database connection (sqlite connection)
        :cursor: Database cursor (sqlite cursor)


    .. _CFApplication: cf.app.CFApplication.html
    """

    def __init__(self, app):
        """
        The constructor of this class takes one argument:

        :Parameters:
            app
                `CFApplication`_ instance

        .. _CFApplication: cf.app.CFApplication.html
        """
        self.app = app
        self.conn = None
        self._user_db = self.app.config.get("userdb.file", USER_DB)
        self.__gobject_init__()
        self.__init_userdb()

    def _get_connection(self):
        return self.conn
    connection = property(fget=_get_connection)

    def __init_userdb(self):
        """Initializes the connection and cursor.

        This method creates the database if necessary.
        """
        log.debug("Initializing user database")
        create = not os.path.isfile(self._user_db)
        self.conn = sqlite.connect(self._user_db)
        self.cursor = self.conn.cursor()
        if create:
            self.__create_userdb()

    def __create_userdb(self):
        """Creates the sqlite3 file and some core tables."""
        sql = "create table sy_table_version (id integer primary key, \
        tablename text, version text)"
        self.cursor.execute(sql)
        from cf import datasources
        datasources.check_userdb(self)

    def get_cursor(self):
        """Returns a new DB-API compliant cursor.

        :Returns: DB-API compliant cursor
        """
        return self.conn.cursor()

    def get_table_version(self, table_name):
        """Returns the version of a table or ``None``.

        ``None`` is returned, if the table isn't registered
        with the `create_table` method, that means, if it's not
        found in ``sy_table_version``.

        :Parameter:
            table_name
                Name of the table

        :Returns: Version of the table, or ``None``
        """
        sql = "select version from sy_table_version \
        where tablename=?"
        try:
            self.cursor.execute(sql, (table_name,))
        except sqlite.OperationalError, e:
            log.warning("sy_table_version not found: %s" % str(e))
            return None
        result = self.cursor.fetchone()
        if result:
            return result[0]
        else:
            return None

    def create_table(self, name, version, statement):
        """Creates and registers a table.

        :Parameter:
            name
                Table name
            version
                Table version
            statement
                ``CREATE`` statement to create the table

        :Returns: ``True`` if the table was successfully created, otherwise ``False``
        """
        if self.get_table_version(name):
            log.error("Table '%s' already exists.", name)
            return False
        try:
            self.cursor.execute(statement)
        except sqlite.OperationalError, e:
            log.error("Failed to create table '%s': %s", name, str(e))
            return False
        sql = "insert into sy_table_version (tablename, version) \
        values (?,?)"
        self.cursor.execute(sql, (name, version))
        self.conn.commit()
        return True

    def drop_table(self, name):
        """Deletes and unregisters a table.

        :Parameter:
            name
                Table name

        :Returns: ``True`` if the table was successfully dropped, otherwise ``False``
        """
        sql = "delete from sy_table_version where tablename=?"
        self.cursor.execute(sql, (name,))
        self.conn.commit()
        sql = "drop table %s" % name
        try:
            self.cursor.execute(sql)
        except sqlite.OperationalError, e:
            log.error("Failed to delete table: %s", str(e))
            return False
        return True



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""Datasource handling"""

import cPickle
import logging

import gobject
import gnomekeyring


log = logging.getLogger("DS")


class DatasourceInfo(gobject.GObject):

    def __init__(self, app, backend, name=None, description=None,
                 options={}, db_id=None):
        """Constructor.

        The constructor of this class takes up to five arguments:

        :Parameter:
            app
                `CFApplication`_ instance
            backend
                Instance of a `DBBackendPlugin`_ plugin
            name
                Datasource name (optional)
            description
                Datasource description (optional)
            options
                Backend specific options (optional)
            db_id
                Databse ID (optional)

        .. _CFApplication: cf.app.CFApplication.html
        .. _DBBackendPlugin: cf.plugins.core.DBBackendPlugin.html
        """
        self.__gobject_init__()
        self.app = app
        self.db_id = db_id
        self.backend = backend
        self.name = name
        self.description = description
        self.options = self._init_options(options)
        self.has_details = False
        self.__conncount = 1L
        self.__connections = list()
        self.internal_connection = None

    def __cmp__(self, other):
        return cmp(self.db_id, other.db_id)

    def _init_options(self, options):
        """Initialize and return options."""
        if self.backend.password_option not in options:
            pwd = self.get_password(self.db_id)
            options[self.backend.password_option] = pwd
        elif self.backend.password_option \
        and self.backend.password_option in options:
            # upgrade from 0.2 -> 0.3: store password in keyring
            gobject.idle_add(self.save, options)
        return options

    def get_label(self):
        return self.backend.get_label(self)

    def save(self, options=None, emit_signal=True):
        cur = self.app.userdb.cursor
        conn = self.app.userdb.conn
        if options is None:
            options = self.options
        if self.backend.password_option \
        and self.backend.password_option in options:
            passwd = options.pop(self.backend.password_option)
        else:
            passwd = None
        if not self.db_id:
            sql = ("insert into datasource (name, description, "
                   "backend, options) values (?,?,?,?)")
            cur.execute(sql, (self.name, self.description, self.backend.id,
                              self.serialize_options(options)))
            self.db_id = cur.lastrowid
            conn.commit()
            signal = "datasource-added"
        else:
            sql = ("update datasource set name=?, description=?, "
                   "backend=?, options=? where id=?")
            cur.execute(sql, (self.name, self.description, self.backend.id,
                              self.serialize_options(options), self.db_id))
            conn.commit()
            signal = "datasource-modified"
        self.store_password(passwd, self.db_id)
        if self.backend.password_option:
            self.options[self.backend.password_option] = passwd
        if emit_signal:
            self.app.datasources.emit(signal, self)

    def delete(self):
        cur = self.app.userdb.cursor
        conn = self.app.userdb.conn
        sql = "delete from datasource where id=?"
        cur.execute(sql, (self.db_id,))
        conn.commit()
        self.store_password(None, self.db_id)
        self.app.datasources.emit("datasource-deleted", self)

    def serialize_options(self, options):
        return cPickle.dumps(options)

    def deserialize_options(self, data):
        return cPickle.loads(data)

    def store_password(self, pwd, db_id):
        """Update password in keyring.

        Args:
          pwd: Password (delete existing if None)
          db_id: Datasource ID
        """
        item_type = gnomekeyring.ITEM_GENERIC_SECRET
        attrs = {"crunchyfrog": db_id}
        try:
            entry = gnomekeyring.find_items_sync(item_type, attrs)
            entry = entry[0]
        except gnomekeyring.NoMatchError:
            entry = None
        if pwd is not None:
            gnomekeyring.item_create_sync(None, item_type,
                                          ("Password for %s"
                                           % self.get_label()),
                                          attrs, pwd, True)
        elif entry is not None:
            gnomekeyring.item_delete_sync(None, entry.item_id)

    def get_password(self, db_id):
        """Get password from keyring.

        Args:
          db_id: Datasource ID

        Returns:
          Password as string or None.
        """
        if db_id is None:
            return None
        item_type = gnomekeyring.ITEM_GENERIC_SECRET
        attrs = {"crunchyfrog": db_id}
        try:
            entry = gnomekeyring.find_items_sync(item_type, attrs)
            return entry[0].secret
        except gnomekeyring.NoMatchError:
            return None

    @classmethod
    def load(cls, app, db_id):
        cur = app.userdb.cursor
        sql = ("select name, description, backend, options "
               "from datasource where id=?")
        cur.execute(sql, (db_id,))
        res = cur.fetchone()
        backend = app.plugins.by_id(res[2], False)
        opts = cPickle.loads(str(res[3]))
        i = cls(app, backend, res[0], res[1], opts, db_id=db_id)
        return i

    @classmethod
    def load_all(cls, app):
        cur = app.userdb.cursor
        sql = "select id from datasource"
        cur.execute(sql)
        r = list()
        for item in cur.fetchall():
            r.append(cls.load(app, item[0]))
        return r

    def get_connections(self):
        return self.__connections

    def add_connection(self, connection):
        connection.connect("closed", self.on_connection_closed)
        self.__connections.append(connection)
        self.app.datasources.emit("datasource-modified", self)

    def on_connection_closed(self, connection):
        if connection in self.__connections:
            self.__connections.remove(connection)
        if connection == self.internal_connection:
            self.internal_connection = None
        gobject.idle_add(self.app.datasources.emit,
                         "datasource-modified", self)

    def dbconnect(self):
        conn = self.backend.dbconnect(self.options)
        conn.datasource_info = self
        conn.conn_number = self.__conncount
        self.__conncount += 1
        if not self.internal_connection:
            self.internal_connection = conn
        self.add_connection(conn)
        return conn

    def dbdisconnect(self):
        while self.__connections:
            conn = self.__connections[0]
            log.info("Closing connection %s" % conn)
            conn.close()


class DatasourceManager(gobject.GObject):
    """Datasource manager

    An instance of this class is accessible through the ``datasources``
    attribute of an `CFApplication`_ instance.

    Signals
    =======
        datasource-added
            ``def callback(manager, datasource_info, user_param1, ...)``

            Emitted when a datsource was added

        datasource-deleted
            ``def callback(manager, datasource_info, user_param1, ...)``

            Emitted when a datasource was removed

        datasource-modified
            ``def callback(manager, datasource_info, user_param1, ...)``

            Emitted when a datasource was modified

    .. _CFApplication: cf.app.CFApplication.html

    """

    __gsignals__ = {
        "datasource-added" : (gobject.SIGNAL_RUN_LAST,
                              gobject.TYPE_NONE,
                              (gobject.TYPE_PYOBJECT,)),
        "datasource-deleted" : (gobject.SIGNAL_RUN_LAST,
                              gobject.TYPE_NONE,
                              (gobject.TYPE_PYOBJECT,)),
        "datasource-modified" : (gobject.SIGNAL_RUN_LAST,
                              gobject.TYPE_NONE,
                              (gobject.TYPE_PYOBJECT,)),
    }

    def __init__(self, app):
        """
        The constructor of this class takes one argument:

        :Parameter:
            app
                `CFApplication`_ instance

        .. _CFApplication: cf.app.CFApplication.html
        """
        self.app = app
        self._cache = list()
        self.__gobject_init__()
        self.connect("datasource-added", self.on_datasource_added)
        self.connect("datasource-deleted", self.on_datasource_deleted)
        self.connect("datasource-modified", self.on_datasource_modified)
        self.app.plugins.connect("plugin-active", self.on_plugin_active)
        for item in DatasourceInfo.load_all(self.app):
            if not item.backend \
            or not self.app.plugins.is_active(item.backend):
                continue
            self._cache.append(item)

    def on_datasource_added(self, manager, datasource_info):
        self._cache.append(datasource_info)

    def on_datasource_deleted(self, manager, datasource_info):
        if datasource_info in self._cache:
            self._cache.remove(datasource_info)

    def on_datasource_modified(self, manager, datasource_info):
        for ds in self._cache:
            if ds != datasource_info:
                continue
            ds.backend = datasource_info.backend
            ds.description = datasource_info.description
            ds.name = datasource_info.name
            ds.options = datasource_info.options

    def on_plugin_active(self, manager, plugin, active):
        from cf.plugins.core import PLUGIN_TYPE_BACKEND
        if not plugin.plugin_type == PLUGIN_TYPE_BACKEND:
            return
        for item in DatasourceInfo.load_all(self.app):
            if item.backend and active and item not in self._cache:
                self.emit("datasource-added", item)
            elif not active and item in self._cache:
                item.dbdisconnect()
                self.emit("datasource-deleted", item)

    def get_all(self):
        """Returns all datasources

        :Returns: List of `DatasourceInfo`_ instances

        .. _DatasourceInfo: cf.datasources.DatasourceInfo.html
        """
        return self._cache


def check_userdb(userdb):
    if not userdb.get_table_version("datasource"):
        sql = ("create table datasource (id integer primary key, "
               "name text, description text, backend text, options text, "
               "last_accessed real, num_accessed integer)")
        userdb.create_table("datasource", "0.1", sql)



# Copyright (C) 2008 Andi Albrecht, albrecht.andi@gmail.com
#
# This module is part of python-sqlparse and is released under
# the BSD License: http://www.opensource.org/licenses/bsd-license.php.

"""This module contains the Query class."""

import logging

from dialects import DialectDefault, Dialect
from errors import SQLParseMultiStatementError
import tokenizer


class Query(object):
    """Interface to a SQL statement.

    The Query class is heavily inspired by the GQL class found in
    google.appengine.ext.gql package.
    """

    SELECT = 'SELECT'
    INSERT = 'INSERT'
    UPDATE = 'UPDATE'
    DELETE = 'DELETE'
    CREATE = 'CREATE'
    DROP = 'DROP'
    OTHER = 'OTHER'

    def __init__(self, statement, dialect=None):
        """Constructor.

        Args:
            statement: SQL statement as string.
            dialect: A Dialect instance.

        Raises:
            AssertionError if statement is not a string or unicode.
            SQLParseMultiStatementError if statement contains
                than one statement.
        """
        assert isinstance(statement, basestring)
        if dialect is None:
            dialect = DialectDefault()
        assert isinstance(dialect, Dialect)
        self._statement = statement
        self._dialect = dialect
        self._tokens = tokenizer._tokenize(statement, self._dialect)
        logging.debug('Query()._tokens --> %r', self._tokens)
        if tokenizer._is_multistatement(self._tokens, self._dialect):
            raise SQLParseMultiStatementError

    def __str__(self):
        return self.join()

    def __repr__(self):
        repr = self._statement[:10]
        if len(self._statement) > 10:
            repr += '...'
        return '<Query \'%s\'>' % repr

    def join(self):
        """Returns a string representation of the query."""
        return tokenizer._join_tokens(self._tokens)

    @property
    def tokens(self):
        return self._tokens

    @property
    def query_type(self):
        """Returns the type of the query.

        Returns:
            One of Query.SELECT, Query.INSERT, Query.UPDATE, Query.DELETE,
                Query.CREATE, Query.DROP, Query.OTHER.
        """
        for token in self._tokens:
            # Ignore trailing whitespaces.
            if token.type == token.WHITESPACE:
                continue
            token = token.upper()
            if token == 'SELECT':
                return Query.SELECT
            elif token == 'INSERT':
                return Query.INSERT
            elif token == 'UPDATE':
                return Query.UPDATE
            elif token == 'DELETE':
                return Query.DELETE
            elif token == 'CREATE':
                return Query.CREATE
            elif token == 'DROP':
                return Query.DROP
            else:
                return Query.OTHER

    def token_start(self):
        """Get the first token."""
        return self.tokens[0]

    def token_next(self, token):
        """Get the next token relative to token."""
        return token._next

    def token_prev(self, token):
        """Get the previous token relative to token."""
        return token._prev

    def token_end(self):
        """Get the last token."""
        return self.tokens[-1]

    def find_keyword(self, keyword, start=None, end=None):
        """Find a keyword token between start and end."""
        if start is None:
            start = self.token_start()
        if end is None:
            end = self.token_end()
        token = start
        while token:
            if token.type == token.KEYWORD \
            and token.upper() == keyword.upper():
                return token
            if token == end:
                break
            token = self.token_next(token)
        return None



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""MS-SQL backend"""

from cf.backends import DBConnectError
from cf.backends.dbapi2helper import DbAPI2Connection
from cf.backends.schema import *
from cf.datasources import DatasourceInfo
from cf.plugins.core import DBBackendPlugin

import time

from gettext import gettext as _

import logging
log = logging.getLogger("MS-SQL")


class MsSQLBackend(DBBackendPlugin):

    id = "crunchyfrog.backend.mssql"
    name = _(u"MS-SQL Plugin")
    description = _(u"Provides access to MS-SQL databases")
    password_option = "password"

    def __init__(self, app):
        DBBackendPlugin.__init__(self, app)
        log.info("Activating MS-SQL backend")
        self.schema = MsSQLSchema()

    def _get_conn_opts_from_opts(self, opts):
        conn_opts = dict()
        if opts["database"]: conn_opts["database"] = opts["database"]
        if opts["host"]: conn_opts["host"] = opts["host"]
        if opts["user"]: conn_opts["user"] = opts["user"]
        if opts["password"]: conn_opts["password"] = opts["password"]
        return conn_opts

    def shutdown(self):
        log.info("Shutting down MS-SQL backend")

    @classmethod
    def get_datasource_options_widgets(cls, data_widgets, initial_data=None):
        return data_widgets, ["database", "host", "user", "password"]

    @classmethod
    def get_label(cls, datasource_info):
        if not datasource_info.options.get("host", None):
            datasource_info.options["host"] = "localhost"
        return DBBackendPlugin.get_label(datasource_info)

    def dbconnect(self, data):
        opts = self._get_conn_opts_from_opts(data)
        if data.get("ask_for_password", False):
            pwd = self.password_prompt()
            if not pwd:
                raise DBConnectError(_(u"No password given."))
            opts["password"] = pwd
        try:
            real_conn = pymssql.connect(**opts)
        except StandardError, err:
            raise DBConnectError(str(err))
        conn = MsSQLConnection(self, self.app, real_conn)
        conn.threadsafety = pymssql.threadsafety
        return conn

    def test_connection(self, data):
        try:
            conn = self.dbconnect(data)
            conn.close()
        except DBConnectError, err:
            return str(err)
        return None

class MsSQLConnection(DbAPI2Connection):

    def __init__(self, *args, **kw):
        DbAPI2Connection.__init__(self, *args, **kw)
        self.coding_hint = "latin-1"

    def get_server_info(self):
        cur = self.cursor()
        cur.execute("SELECT @@VERSION")
        ver = cur.fetchall()[0][0]
        ver = ver.splitlines()[0].strip()+" - "+ver.splitlines()[-1].strip()
        cur.close()
        return ver

class MsSQLSchema(SchemaProvider):

    def q(self, connection, sql):
        cur = connection.cursor()._cur
        cur.execute(sql)
        return cur.fetchall()

    def fetch_children(self, connection, parent):
        if isinstance(parent, DatasourceInfo):
            return [SchemaCollection()]

        elif isinstance(parent, SchemaCollection):
            sql = "select catalog_name from information_schema.schemata"
            return [Schema(item[0]) for item in self.q(connection, sql)]

        elif isinstance(parent, Schema):
            return [TableCollection(schema=parent),
                    ViewCollection(schema=parent)]

        elif isinstance(parent, TableCollection):
            schema = parent.get_data("schema")
            sql = "select table_name from information_schema.tables \
            where table_type = 'BASE TABLE' \
            and table_catalog = '%s'" % schema.name
            return[Table(item[0], schema=schema)
                   for item in self.q(connection, sql)]

        elif isinstance(parent, ViewCollection):
            schema = parent.get_data("schema")
            sql = "select table_name from information_schema.tables \
            where table_type = 'VIEW' \
            and table_catalog = '%s'" % schema.name
            return[View(item[0], schema=schema)
                   for item in self.q(connection, sql)]

        elif isinstance(parent, (Table, View)):
            return [ColumnCollection(table=parent)]

        elif isinstance(parent, ColumnCollection):
            table = parent.get_data("table")
            schema = table.get_data("schema")
            sql = "select column_name from information_schema.columns \
            where table_catalog = '%s' and table_name = '%s'"
            sql = sql % (schema.name, table.name)
            return [Column(item[0], table=table)
                    for item in self.q(connection, sql)]


MsSQLBackend.INIT_ERROR = None
try:
    import pymssql
except ImportError:
    MsSQLBackend.INIT_ERROR = _(u"Python module pymssql not found.")



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id$

"""Instance class"""

import gtk
import gtk.glade
import gnome
import gnomevfs
import gobject

import os

from gettext import gettext as _

from cf import release
from cf.ui import GladeWidget
from cf.ui import pdock
from cf.ui.browser import Browser
from cf.ui.datasources import DatasourceManager
from cf.ui.editor import Editor, EditorWindow
from cf.ui.queries import QueriesNotebook
from cf.ui.statusbar import CFStatusbar
from cf.ui.toolbar import CFToolbar
from cf.ui.widgets import ConnectionButton
from cf.ui.confirmsave import ConfirmSaveDialog

class CFInstance(GladeWidget):

    def __init__(self, app):
        """
        The constructor of this class takes one argument:

        :Parameter:
            app
                `CFApplication`_ instance

        .. _CFApplication: cf.app.CFApplication.html
        """
        GladeWidget.__init__(self, app, "crunchyfrog", "mainwindow")
        self._editor = None
        self._editor_conn_tag = None
        self._editors = list()

    def _setup_widget(self):
        # Window state
        if self.app.config.get("gui.width", -1) != -1:
            self.widget.resize(self.app.config.get("gui.width"),
                               self.app.config.get("gui.height"))
        if self.app.config.get("gui.maximized", False):
            self.widget.maximize()
        # Tooltips
        self.tt = gtk.Tooltips()
        # Toolbar
        self.toolbar = CFToolbar(self)
        self.toolbar.show_all()
        # Statusbar
        self.statusbar = CFStatusbar(self.app, self.xml)
        # Dock
        self.dock = pdock.Dock()
        box = self.xml.get_widget("box_main")
        box.pack_start(self.dock, True, True)
        box.reorder_child(self.dock, 2)
        self.dock.show_all()
        # Queries
        self.queries = QueriesNotebook(self.app, self)
        item = pdock.DockItem(self.dock, "queries", self.queries, _(u"Queries"),
                              "gtk-edit", None, pdock.DOCK_ITEM_BEH_LOCKED)
        self.dock.add_item(item)
        # Browser
        self.browser = Browser(self.app, self)
        item = pdock.DockItem(self.dock, "browser", self.browser, _(u"Navigator"),
                              "gtk-find", gtk.POS_LEFT, pdock.DOCK_ITEM_BEH_CANT_CLOSE)
        self.dock.add_item(item)
        self.browser.set_data("dock_item", item)
        gobject.idle_add(self.xml.get_widget("mn_navigator").set_active, self.app.config.get("navigator.visible", True))
        # Menu bar
        recent_menu = gtk.RecentChooserMenu(self.app.recent_manager)
        filter = gtk.RecentFilter()
        filter.add_mime_type("text/x-sql")
        recent_menu.add_filter(filter)
        recent_menu.set_filter(filter)
        recent_menu.set_show_not_found(False)
        recent_menu.set_sort_type(gtk.RECENT_SORT_MRU)
        recent_menu.connect("item-activated", self.on_recent_item_activated)
        recent_item = self.xml.get_widget("mn_recent")
        recent_item.set_submenu(recent_menu)

    def _init_ui(self, argv):
        pass

    def on_about(self, *args):
        def open_url(dialog, url):
            gnome.url_show(url)
        gtk.about_dialog_set_url_hook(open_url)
        dlg = gtk.AboutDialog()
        dlg.set_name(release.name)
        dlg.set_version(release.version)
        dlg.set_copyright(release.copyright)
        dlg.set_license(release.license)
        dlg.set_website(release.url)
        dlg.set_website_label(release.url)
        dlg.set_logo_icon_name(release.appname)
        dlg.set_program_name(release.appname)
        dlg.set_translator_credits(release.translators)
        dlg.run()
        dlg.destroy()

    def on_commit(self, *args):
        if not self._editor:
            return
        gobject.idle_add(self._editor.commit)

    def on_rollback(self, *args):
        if not self._editor:
            return
        gobject.idle_add(self._editor.rollback)

    def on_begin_transaction(self, *args):
        if not self._editor:
            return
        gobject.idle_add(self._editor.begin_transaction)

    def on_configure_event(self, win, event):
        config = self.app.config
        if not config.get("gui.maximized"):
            config.set("gui.width", event.width)
            config.set("gui.height", event.height)

    def _get_clipboard(self):
        display = gtk.gdk.display_manager_get().get_default_display()
        return gtk.Clipboard(display, "CLIPBOARD")

    def on_copy(self, *args):
        if not self._editor:
            return
        self._editor.textview.get_buffer().copy_clipboard(self._get_clipboard())

    def on_paste(self, *args):
        if not self._editor:
            return
        self._editor.textview.get_buffer().paste_clipboard(self._get_clipboard(), None, True)

    def on_cut(self, *args):
        if not self._editor:
            return
        self._editor.textview.get_buffer().cut_clipboard(self._get_clipboard(), True)

    def on_recent_item_activated(self, chooser):
        self.new_editor(chooser.get_current_uri())

    def on_delete(self, *args):
        if not self._editor:
            return
        self._editor.textview.get_buffer().delete_selection(True, True)

    def on_datasource_manager(self, *args):
        dlg = DatasourceManager(self.app, self)
        dlg.run()
        dlg.destroy()

    def on_editor_connection_changed(self, editor, connection):
        if connection:
            self.set_title(connection.get_label()+" - CrunchyFrog")
        else:
            self.set_title("CrunchyFrog")

    def on_execute_query(self, *args):
        gobject.idle_add(self._editor.execute_query)

    def on_help(self, *args):
        self.show_help()

    def on_navigator_toggled(self, menuitem):
        if menuitem.get_active():
            self.browser.get_data("dock_item").show()
            self.app.config.set("navigator.visible", True)
        else:
            self.browser.get_data("dock_item").hide()
            self.app.config.set("navigator.visible", False)

    def on_new_instance(self, *args):
        self.app.new_instance(tuple())

    def on_open_file(self, *args):
        dlg = gtk.FileChooserDialog(_(u"Select file"),
                            self.widget,
                            gtk.FILE_CHOOSER_ACTION_OPEN,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dlg.set_current_folder(self.app.config.get("editor.recent_folder", ""))
        filter = gtk.FileFilter()
        filter.set_name(_(u"All files (*)"))
        filter.add_pattern("*")
        dlg.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name(_(u"SQL files (*.sql)"))
        filter.add_pattern("*.sql")
        dlg.add_filter(filter)
        dlg.set_filter(filter)
        recent_chooser = gtk.RecentChooserWidget(self.app.recent_manager)
        filter = gtk.RecentFilter()
        filter.add_mime_type("text/x-sql")
        filter.set_name(_(u"SQL files (*.sql)"))
        recent_chooser.add_filter(filter)
        recent_chooser.set_filter(filter)
        recent_chooser.set_show_not_found(False)
        recent_chooser.set_sort_type(gtk.RECENT_SORT_MRU)
        recent_chooser.set_show_icons(True)
        recent_chooser.set_show_tips(True)
        def dlg_set_uri(chooser, dlg):
            if chooser.get_current_uri():
                dlg.set_uri(chooser.get_current_uri())
        recent_chooser.connect("selection-changed", dlg_set_uri, dlg)
        recent_chooser.connect("item-activated", lambda chooser: dlg.response(gtk.RESPONSE_OK))
        exp = gtk.Expander(_(u"_Recent files:"))
        exp.add(recent_chooser)
        exp.set_use_underline(True)
        dlg.set_extra_widget(exp)
        recent_chooser.show()
        if dlg.run() == gtk.RESPONSE_OK:
            self.new_editor(dlg.get_filename())
            self.app.config.set("editor.recent_folder", dlg.get_current_folder())
        dlg.destroy()

    def on_save_file(self, *args):
        if not self._editor:
            return
        self._editor.save_file()

    def on_save_file_as(self, *args):
        if not self._editor:
            return
        self._editor.save_file_as()

    def on_report_problem(self, *args):
        gobject.idle_add(self.open_website, "http://code.google.com/p/crunchyfrog/issues/list")

    def on_open_devpages(self, *args):
        gobject.idle_add(self.open_website, "http://cf.andialbrecht.de")

    def on_open_helptranslate(self, *args):
        gobject.idle_add(self.open_website, "https://translations.launchpad.net/crunchyfrog/trunk/")

    def on_preferences(self, *args):
        self.app.preferences_show()

    def on_plugins_configure(self, *args):
        self.app.preferences_show("plugins")

    def on_query_new(self, *args):
        self.new_editor()

    def on_quit(self, *args):
        if not self.check_unsaved_changes():
            return
        self.widget.destroy()

    def on_window_state_event(self, win, event):
        config = self.app.config
        bit = gtk.gdk.WINDOW_STATE_MAXIMIZED.value_names[0] in event.new_window_state.value_names
        config.set("gui.maximized", bit)

    def new_editor(self, fname=None):
        """Creates a new SQL editor

        :Parameter:
            fname
                If given, the file ``fname`` is opened with this editor
        :Returns: `Editor`_ instance

        .. _Editor: cf.ui.editor.Editor.html
        """
        editor = Editor(self.app, self)
        if fname:
            if fname.startswith("file://"):
                fname = gnomevfs.get_local_path_from_uri(fname)
            editor.set_filename(fname)
        if self.app.config.get("editor.open_in_window"):
            editor.show_in_separate_window()
        else:
            self.queries.attach(editor)
        editor.show_all()
        self._editors.append(editor)
        return editor

    def check_unsaved_changes(self):
        changed_editors = list()
        for editor in self._editors:
            if editor.contents_changed():
                changed_editors.append(editor)
        if changed_editors:
            dlg = ConfirmSaveDialog(self, changed_editors)
            resp = dlg.run()
            if resp == 1:
                ret = dlg.save_files()
            elif resp == 2:
                ret = True
            else:
                ret = False
            dlg.destroy()
        else:
            ret = True
        return ret

    def open_website(self, url):
        gnome.url_show(url)

    def set_editor_active(self, editor, active):
        if not active:
            editor = None
        if self._editor_conn_tag and self._editor:
            self._editor.disconnect(self._editor_conn_tag)
            self._editor_conn_tag = None
        self._editor = editor
        if self._editor:
            self._editor_conn_tag = self._editor.connect("connection-changed", self.on_editor_connection_changed)
            self.on_editor_connection_changed(self._editor, self._editor.connection)
        else:
            self.set_title("CrunchyFrog")
        self.toolbar.set_editor(editor)
        self.app.plugins.editor_notify(editor, self)

    def get_editor(self):
        """Returns the active editor

        :Returns: editor instance or ``None``
        """
        return self._editor

    def show_help(self, topic=None):
        gnome.help_display(release.appname, topic)


class InstanceSelector(GladeWidget):

    def __init__(self, client):
        self.client = client
        GladeWidget.__init__(self, None, "crunchyfrog", "instanceselector")

    def _setup_widget(self):
        model = gtk.ListStore(int, str)
        self.list = self.xml.get_widget("list_instances")
        self.list.set_model(model)
        col = gtk.TreeViewColumn("", gtk.CellRendererText(), text=1)
        self.list.append_column(col)
        for id, title in self.client.get_instances():
            model.append([id, title])

    def _setup_connections(self):
        sel = self.list.get_selection()
        sel.connect("changed", self.on_isel_changed)
        self.xml.get_widget("btn_newinstance").connect("toggled", self.on_newi_toggled)

    def on_newi_toggled(self, btn):
        sel = self.list.get_selection()
        model = self.list.get_model()
        if btn.get_active():
            sel.unselect_all()
        else:
            sel.select_iter(model.get_iter_first())
        self.list.set_sensitive(not btn.get_active())

    def on_isel_changed(self, selection):
        model, iter = selection.get_selected()
        if not iter:
            self.xml.get_widget("btn_newinstance").set_active(True)
        else:
            self.xml.get_widget("btn_activeinstance").set_active(True)

    def get_instance_id(self):
        if self.xml.get_widget("btn_newinstance").get_active():
            return None
        else:
            sel = self.list.get_selection()
            model, iter = sel.get_selected()
            if iter:
                return model.get_value(iter, 0)
            return None
